# import os
import csv
# import numpy as np

import twobitreader as tbr

from collections import defaultdict
from collections import namedtuple
# from itertools import cycle
from tqdm import tqdm

from mutagene.dna import nucleotides, complementary_nucleotide
from mutagene.dna import codon_table
# from mutagene.dna import chromosome_name_mapping

import logging
logger = logging.getLogger(__name__)


def read_protein_mutations_MAF_file(fname, genome, motifs=False):
    with open(fname) as infile:
        return read_protein_mutations_MAF(infile, genome)


def read_protein_mutations_MAF(infile, genome, motifs=False):
    mutations = defaultdict(dict)
    processing_stats = {'loaded': 0, 'skipped': 0, 'nsamples': 0, 'format': 'unknown'}

    if not infile:
        logger.warning("No input file")
        return mutations, processing_stats

    if genome.upper() != 'MAF':
        fname = genome if genome.endswith('.2bit') else genome + '.2bit'
        twobit = tbr.TwoBitFile(fname)

    try:
        reader = csv.reader((row for row in infile if not row.startswith('#')), delimiter='\t')
        # get names from column headers
        header = next(reader)
        header = tuple(map(lambda s: s.lower().replace('.', '_'), header))
        # print(header)
        MAF = namedtuple("MAF", header, rename=True)
    except ValueError:
        raise
        logger.warning("MAF format not recognized")
        return mutations, processing_stats

    N_loaded = N_skipped = 0

    for data in tqdm(map(MAF._make, reader), leave=False):
        # print(data)
        try:
            if hasattr(data, 'tumor_sample_barcode'):
                sample = data.tumor_sample_barcode
            elif hasattr(data, 'sample_id'):
                sample = data.sample_id
            else:
                raise ValueError("Sample ID is not defined in MAF file")

            if hasattr(data, 'transcript_id'):
                transcript = data.transcript_id
            elif hasattr(data, 'annotation_transcript'):
                transcript = data.annotation_transcript
            else:
                logger.debug("Transcript undefined")
                N_skipped += 1
                continue

            HGVSc = None
            if hasattr(data, 'cdna_change'):
                HGVSc = data.cdna_change
            if hasattr(data, 'txchange'):
                HGVSc = data.txchange
            if hasattr(data, 'hgvsc'):
                HGVSc = data.hgvsc

            if HGVSc is None:
                logger.warning("Could not find cDNA_Change or TxChange or HGVSc fields in MAF file")
                break

            if ':' in HGVSc:
                HGVSc = HGVSc.split(':')[1]

            if not HGVSc.startswith("c."):
                N_skipped += 1
                logger.debug("Skip HGVSc c.: " + HGVSc)
                continue

            unexpected_symbols = ['_', 'del', 'dup', '-', '+']
            try:
                for us in unexpected_symbols:
                    if us in HGVSc:
                        N_skipped += 1
                        logger.debug("Skip indel, dup or splice " + HGVSc)
                        raise ValueError()
            except ValueError:
                continue

            if '>' in HGVSc:
                # c.368C>T
                cDNA_position = int(HGVSc.split('.')[1][:-3])
            else:
                # c.C232T
                cDNA_position = int(HGVSc.split('.')[1][1:-1])
            offset = ((cDNA_position % 3) - 1) % 3  # determine codon offset: 0, 1, 2

            HGVSp = None
            if hasattr(data, 'protein_change'):
                HGVSp = data.protein_change
            if hasattr(data, 'hgvsp_short'):
                HGVSp = data.hgvsp_short
            if hasattr(data, 'aachange'):
                HGVSp = data.aachange

            if HGVSp is None:
                logger.warning("Could not find HGVSp_Short or Protein_Change or AAChange fields in MAF file")
                break

            if not HGVSp.startswith("p."):
                N_skipped += 1
                logger.debug("Skip HGVSp p.: " + HGVSp)
                continue

            if '_' in HGVSp or HGVSp.endswith('fs'):
                N_skipped += 1
                logger.debug("Skip fs HGVSp: " + HGVSp)
                continue


            protein_mutation = HGVSp.split('.')[1].upper()
            P = protein_mutation[0]
            Q = protein_mutation[-1]

            if not hasattr(data, 'hugo_symbol'):
                logger.warning("Could not find Hugo_Symbol in MAF file")
                break

            ###### Nucleotide stuff ######
            if not hasattr(data, 'variant_classification'):
                logger.warning("Could not find Variant_Classification in MAF file")
                break

            if data.variant_classification.lower() not in ("nonsense_mutation", "missense_mutation", "silent"):
                N_skipped += 1
                logger.debug("Variant_Classification")
                continue

            chrom = None
            start = None

            if hasattr(data, 'chromosome'):
                chrom = data.chromosome
            if hasattr(data, 'start_position'):
                start = int(data.start_position)

            if not chrom or not start:
                logger.warning("Could not find Chromosome and Start_Position or Start_position in MAF file")
                break

            chromosome = chrom if chrom.startswith('chr') else 'chr' + chrom

            if not motifs and hasattr(data, 'ref_context'):
                # context bundled in MAF file
                context = data.ref_context.upper()
                start = 10 + 1  # ref_context: 10 + 1 + 10
                seq5_fwd = context[start - offset - 2: start - offset + 3]
                seq5_rev_mirror = context[start - (2 - offset) - 2: start - (2 - offset) + 3]
                seq5_rev = "".join([complementary_nucleotide[x] for x in reversed(seq5_rev_mirror)])
            else:
                if genome.upper() == 'MAF':
                    logger.warning("ref_context not found in MAF file. Provide genome name argument -g hg19, hg38, mm10, see http://hgdownload.cse.ucsc.edu/downloads.html for more")
                    break

                try:
                    seq5_fwd = twobit[chromosome][start - offset - 2: start - offset + 3].upper()
                    seq5_rev_mirror = twobit[chromosome][start - (2 - offset) - 2: start - (2 - offset) + 3].upper()
                except ValueError:
                    # could not read reference genome with given coordinates
                    logger.debug("Detected mismatch of mutated reference allele with the reference genome, check if using correct genome: " + data.Protein_Change)
                    # logger.warning("Chr {} {} {}".format(chromosome, start, offset))
                    N_skipped += 1
                    continue

                if len(set(seq5_fwd + seq5_rev_mirror) - set(nucleotides)) > 0:
                    # detected non-nucleotide letters (including N)
                    N_skipped += 1
                    logger.debug("Detected non-nucleotide letters (including N)")
                    continue
                seq5_rev = "".join([complementary_nucleotide[x] for x in reversed(seq5_rev_mirror)])

            if codon_table[seq5_fwd[1:4]] == P:
                seq5 = seq5_fwd
            elif codon_table[seq5_rev[1:4]] == P:
                seq5 = seq5_rev
            else:
                # print(seq5_fwd[1:4], seq5_rev[1:4], c1)
                N_skipped += 1
                # logger.debug(context[:10] + " " + context[10] + " " + context[11:] + " " + seq5_fwd + " " + seq5_rev + " " + c1 + " " + c2)
                logger.debug("Sequence missmatch of mutation with the genome, check if using the correct genome assembly: " + HGVSp)
                continue

            mutations[sample][(data.hugo_symbol, transcript, protein_mutation)] = {'seq5': seq5}
        except Exception as e:
            N_skipped += 1
            logger.debug("General MAF parsing exception " + str(e))
            continue

    N_loaded = 0
    for sample, sample_mutations in mutations.items():
        N_loaded += int(len(sample_mutations.keys()))

    processing_stats = {
        'loaded': N_loaded,
        'skipped': N_skipped,
        'nsamples': len(mutations.keys()),
        'format': 'MAF'
    }

    gene_transcript_mapping = {}
    flat_mutations = {}
    for sample, sample_mutations in mutations.items():
        for (gene, transcript, protein_mutation), props in sample_mutations.items():
            if gene in gene_transcript_mapping:
                if transcript != gene_transcript_mapping[gene]:
                    continue
            else:
                # use first encountered transcript as canonical for the gene
                # ignore other mutations
                gene_transcript_mapping[gene] = transcript

            if (gene, protein_mutation) not in flat_mutations:
                # note that mutability would be represented only for one nucleotide mutation
                flat_mutations[(gene, protein_mutation)] = {
                    'seq5': {
                        props['seq5']: 1
                    }
                }
            else:
                if props['seq5'] not in flat_mutations[(gene, protein_mutation)]['seq5']:
                    flat_mutations[(gene, protein_mutation)]['seq5'][props['seq5']] = 0
                flat_mutations[(gene, protein_mutation)]['seq5'][props['seq5']] += 1

    return flat_mutations, processing_stats



#
import numpy as np

# from .io import read_profile
# from .io import format_profile
from mutagene.signatures.identify import decompose_mutational_profile_counts


def convert_to_list(name_to_idx, d):
    v = [0] * N
    for value in d:
        idx = name_to_idx.get(value['name'])
        if idx is None:
            # print(value['name'], 'not found in name_to_idx')
            continue
        v[idx] = value['score']
    return v


def get_scores(d):
    ll = 0
    frob = 0
    frob0 = 0
    js = 0
    kl = 0
    for value in d:
        if value['name'] == 'LogLik':
            ll = value['score']
        if value['name'] == 'Frobenius':
            frob = value['score']
        if value['name'] == 'FrobeniusZero':
            frob0 = value['score']
        if value['name'] == 'DivergenceJS':
            js = value['score']
        if value['name'] == 'DivergenceKL':
            kl = value['score']
    return ll, frob, frob0, js, kl


def benchmark_simulated(results_fname, signature_names, W):
    name_to_idx = {}
    for i, s in enumerate(signature_names):
        name_to_idx[s] = i

    W = np.array(W).T
    print(W.shape)
    N = W.shape[1]

    signature_names = list(name_to_idx.keys())
    signature_ids = [str(name.split()[1]) for name in signature_names]

    np.set_printoptions(precision=4)

    with open(results_fname, 'w') as report:
        report.write("TEST\tMUTATIONS\tMETH\tROUND\tSIGNATURE\tVALUE\tSE_E\tMSE_M\tMSE_E\tLL\tFROB\tFROB0\tJS\tKL\n")
        # for j in range(N):
        #     report.write("S{}\t".format(j + 1))
        # report.write("MSE\n")

        for i in range(N):

            for N_mutations in [10, 50, 100, 500, 1000, 10000]:

                for iteration in range(10):
                    h0 = np.zeros(N)
                    h0[i] = int(0.8 * N_mutations)
                    for j in range(N):
                        if i == j:
                            continue

                    # 20% uniform noise
                    # for k in range(int(0.2 * N_mutations)):
                    #     h0[random.randrange(N)] += 1
                    h0 += np.random.multinomial(int(0.2 * N_mutations), [1.0 / N] * N)  # uniform distribution
                    h0_counts = h0.copy()

                    h0 /= h0.sum()
                    v0 = W.dot(h0)
                    v0_counts = np.random.multinomial(N_mutations, v0 / v0.sum())  # np.ceil(v0 * N_mutations)

                    # oname = prefix + "profile"
                    # # print("ONAME:", oname)
                    # with open(oname, 'w') as o:
                    #     query_formatted = format_profile(h0.tolist())
                    #     o.write(query_formatted)

                    # oname = prefix + "counts"
                    # with open(oname, 'w') as o:
                    #     query_formatted = format_profile(h0.tolist(), counts=True)
                    #     o.write(query_formatted)

                    THRESHOLD = 0.0
                    # print(h0_counts)
                    # print(v0_counts)
                    _, _, exposure = decompose_mutational_profile_counts(v0_counts, W, 'MLE', others_threshold=THRESHOLD)

                    h1 = np.array(convert_to_list(name_to_idx, exposure))
                    # h1 += min(1.0 - h1.sum(), 0)
                    h1_counts = np.random.multinomial(N_mutations, h1 / h1.sum())
                    ll, frob, frob0, js, kl = get_scores(exposure)

                    # print(np.round(h0, 3))
                    # print(np.round(h1, 3))
                    v1 = W.dot(h1)

                    # convert exposure to mutational burden
                    # v1_counts = np.ceil(v1 * N_mutations)

                    # convert exposure to mutational burden
                    v1_counts = np.random.multinomial(N_mutations, v1 / v1.sum())

                    MSE_M = np.mean((v0_counts - v1_counts)**2)
                    MSE_E = np.mean((h0 - h1)**2)

                    for j in range(N):
                        report.write("{}_{}_{}\t{}\t{}\t{}\t".format(signature_ids[i], 80, 20, N_mutations, "MLE", iteration + 1))
                        report.write("{}\t{}\t".format(signature_names[j], int(h1_counts[j])))
                        SE_E = (h0[j] - h1[j])**2
                        report.write("{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\n".format(SE_E, MSE_M, MSE_E, ll, frob, frob0, js, kl))

                    print(signature_ids[i], N_mutations, iteration, round(MSE_M, 4), round(MSE_E, 4), round(ll, 4), round(frob, 4), round(frob0, 4), round(js, 4), round(kl, 4))


def benchmark_2combinations(results_fname, signature_names, W):
    name_to_idx = {}
    for i, s in enumerate(signature_names):
        name_to_idx[s] = i

    W = np.array(W).T
    N = W.shape[1]

    signature_names = list(name_to_idx.keys())
    signature_ids = [str(name.split()[1]) for name in signature_names]

    np.set_printoptions(precision=4)

    noise_level = 0.05
    ratio = 0.7
    n_sample = 20

    with open(results_fname, 'w') as report:
        report.write("ratio\tn_sample\tnoise_level\tsignatures\tSIGNATURE\tVALUE\tSE_E\tMSE_M\tMSE_E\tLL\tFROB\tFROB0\tJS\tKL\n")
        # for j in range(N):
        #     report.write("S{}\t".format(j + 1))
        # report.write("MSE\n")

        for i in range(N):

            # for N_mutations in [10, 50, 100, 500, 1000, 10000]:
            for N_mutations in [50, 500]:

                for iteration in range(10):
                    h0 = np.zeros(N)
                    h0[i] = int(0.8 * N_mutations)
                    for j in range(N):
                        if i == j:
                            continue

                    # 20% uniform noise
                    # for k in range(int(0.2 * N_mutations)):
                    #     h0[random.randrange(N)] += 1
                    h0 += np.random.multinomial(int(0.2 * N_mutations), [1.0 / N] * N)  # uniform distribution
                    h0_counts = h0.copy()

                    h0 /= h0.sum()
                    v0 = W.dot(h0)
                    v0_counts = np.random.multinomial(N_mutations, v0 / v0.sum())  # np.ceil(v0 * N_mutations)

                    THRESHOLD = 0.0
                    # print(h0_counts)
                    # print(v0_counts)
                    _, _, exposure = decompose_mutational_profile_counts(v0_counts, W, 'MLE', others_threshold=THRESHOLD)

                    h1 = np.array(convert_to_list(name_to_idx, exposure))
                    # h1 += min(1.0 - h1.sum(), 0)
                    h1_counts = np.random.multinomial(N_mutations, h1 / h1.sum())
                    ll, frob, frob0, js, kl = get_scores(exposure)

                    # print(np.round(h0, 3))
                    # print(np.round(h1, 3))
                    v1 = W.dot(h1)

                    # convert exposure to mutational burden
                    # v1_counts = np.ceil(v1 * N_mutations)

                    # convert exposure to mutational burden
                    v1_counts = np.random.multinomial(N_mutations, v1 / v1.sum())

                    MSE_M = np.mean((v0_counts - v1_counts)**2)
                    MSE_E = np.mean((h0 - h1)**2)

                    for j in range(N):
                        report.write("{}_{}_{}\t{}\t{}\t{}\t".format(signature_ids[i], 80, 20, N_mutations, "MLE", iteration + 1))
                        report.write("{}\t{}\t".format(signature_names[j], int(h1_counts[j])))
                        SE_E = (h0[j] - h1[j])**2
                        report.write("{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\n".format(SE_E, MSE_M, MSE_E, ll, frob, frob0, js, kl))

                    print(signature_ids[i], N_mutations, iteration, round(MSE_M, 4), round(MSE_E, 4), round(ll, 4), round(frob, 4), round(frob0, 4), round(js, 4), round(kl, 4))



import multiprocessing
import json
import requests
from tqdm import tqdm

from mutagene.dna import complementary_nucleotide


def mp_ensembl_worker(raw_mutations_chunk):
    """ Helper function for ENSEMBL API call """

    cn = complementary_nucleotide
    verify_reference = {}

    contexts = {}  # result

    api_input = {'regions': []}
    for i, (assembly, chrom, pos, x, y) in enumerate(raw_mutations_chunk):
        # print("!!!", assembly, chrom, pos, x, y)
        api_input['regions'].append("{}:{}..{}".format(chrom, int(pos) - 1, int(pos) + 1))
        verify_reference[(chrom, int(pos))] = x
    api_input_json = json.dumps(api_input)

    asm = {
        38: "GRCh38",
        37: "GRCh37",
        36: "NCBI36",
        35: "NCBI35",
        34: "NCBI34"}

    server = "https://rest.ensembl.org"
    ext = "/sequence/region/human/?coord_system_version={}".format(asm.get(int(assembly), "GRCh38"))

    r = requests.post(server + ext, headers={
        "Content-Type": "application/json",
        "Accept": "application/json"}, data=api_input_json, timeout=2.000, verify=False)  # 2 second timeout
    if not r.ok:
        print("ENSEMBL Exception")
        return contexts
        # raise Exception("POST request does not work")
    decoded = r.json()

    if 'error' in decoded:
        return contexts

    for i, record in enumerate(decoded):
        if 'id' not in record:
            continue
        _, asm, chrom, begin, end, expand = record['id'].split(":")
        pos = int(begin) + 1
        if _ != "chromosome":
            continue

        seq = record['seq']
        nuc5 = seq[0]
        nuc = seq[1]
        nuc3 = seq[2]
        x = verify_reference.get((chrom, pos))
        if nuc != x:
            if cn[nuc] == x:
                nuc3 = cn[nuc5]
                nuc5 = cn[nuc3]
            else:
                # print("Could not match reference allele in ", raw_mutations_chunk[i], "in sequence", seq)
                nuc3 = 'N'
                nuc5 = 'N'
        # print("{} {}: {} [{} > ?] {}".format(chrom, pos, nuc5, x, nuc3))
        contexts[(chrom, pos)] = (nuc5, nuc3)
    return contexts


def get_context_ensembl(mutations, assembly):
    """
    Use Ensembl to get context of mutations
    ENSEMBL REST API http://rest.ensembl.org/documentation/info/sequence_region_post
    """
    raw_mutations = []
    for (chrom, pos, x, y) in mutations:
        raw_mutations.append((assembly, chrom, pos, x, y))

    L = len(raw_mutations)
    MAX_POST_SIZE = 50  # https://github.com/Ensembl/ensembl-rest/blob/78eebccb589798662c0ea758081ba2b2cb6acbca/ensembl_rest.conf.default     max_post_size = 50

    chunks = []
    for chunk in range(0, L, MAX_POST_SIZE):
        chunks.append(raw_mutations[chunk: chunk + MAX_POST_SIZE])

    try:
        p = multiprocessing.Pool(7)
        cc = p.map(mp_ensembl_worker, tqdm(chunks), MAX_POST_SIZE)
    except:
        # raise
        cc = []
    finally:
        p.close()
        p.join()

    contexts = {}
    for c in cc:
        contexts.update(c)

    if len(contexts) != len(raw_mutations):
        # print("Not all mutations have context")
        pass
    return contexts



import re
import math
import scipy.stats as stats
from statsmodels.stats.multitest import multipletests
import numpy as np
import pandas as pd

from tqdm import tqdm
# import functools
import pprint

from mutagene.dna import (
    nucleotides, complementary_nucleotide,
    bases_dict,
    # comp_dict,
    extended_nucleotides, complementary_extended_nucleotide)

from mutagene.io.motifs import get_known_motifs

import logging
logger = logging.getLogger(__name__)


def identify_motifs(samples_mutations, custom_motif=None, strand=None, threshold=None, dump_matches=None, stat_type=None):
    """
    :param samples_mutations: list of mutations from input file
    :param custom_motif: specified motif to search for
    :param strand: strand(s) to search on (T: transcribed, N: non-transcribed, A: any, or a combination theirof: 'TNA')
    :param dump_matches: pass through to process_mutations, stores all motif matches
    :param stat_type: pass through to process_mutations, choose statistical test
    :return: command-line output
    """
    motif_matches = []
    sig_motif_matches = []
    pvals = []

    if strand is None:
        strand = 'A'
    else:
        strand = set(strand)  # in case TNA codes repeat

    if threshold is None:
        threshold = 0.05

    if custom_motif:
        search_motifs = scanf_motif(custom_motif)
    else:
        motifs = get_known_motifs()
        search_motifs = motifs.copy()
    # search_motifs.extend(scanf_motif(custom_motif))

    _strand_map = {
        'T': 'transcribed',
        'N': 'non-transcribed',
        'A': 'any strand'
    }

    disable_progress_bar = logger.getEffectiveLevel() == logging.DEBUG

    for sample, mutations in tqdm(samples_mutations.items(), leave=False, disable=disable_progress_bar):
        if mutations is not None and len(mutations) > 0:
            first_mut_seq_with_coords = mutations[0][-1]
            window_size = (len(first_mut_seq_with_coords) - 1) // 2

            for m in tqdm(search_motifs, leave=False, disable=disable_progress_bar):
                for s in strand:
                    result, saved_data = process_mutations(
                        mutations,
                        m['motif'],
                        m['position'],
                        m['ref'],
                        m['alt'],
                        window_size,
                        s,
                        stat_type=stat_type)

                    if dump_matches:
                        for chrom, pos in saved_data['mutation_motif']:
                            dump_matches.write(
                                "chr{}\t{}\t{}\t{}\t{}\t{}\n".format(
                                    chrom, pos, int(pos) + 1, sample, m['logo'], _strand_map[s]))

                    debug_data = {
                        'sample': sample,
                        'motif': m['logo'],
                        'strand': s}
                    debug_data.update(result)
                    debug_string = pprint.pformat(debug_data, indent=4)
                    logger.debug(debug_string)

                    motif_matches.append({
                        'sample': sample,
                        'mutagen': m['name'],
                        'motif': m['logo'],
                        'strand': _strand_map[s],
                        'enrichment': result['enrichment'],
                        'mut_min': result['mutation_load'],
                        'mut_max': result['bases_mutated_in_motif'],
                        'odds_ratio': result['odds_ratio'],
                        'pvalue': result['pvalue']
                    })
                    pvals.append(result['pvalue'])

    qvalues = get_corrected_pvalues(pvals)
    for i, motif_dict in enumerate(motif_matches):
        motif_matches[i]['qvalue'] = qvalues[i]
        if motif_dict['mut_min'] == 0:
            continue
        if motif_dict['qvalue'] >= threshold:
            continue
        sig_motif_matches.append(motif_dict)

    return sig_motif_matches


def scanf_motif(custom_motif):
    """ recognize motif syntax like A[C>T]G and create a motif entry """
    m = re.search(
        r'([' + extended_nucleotides + ']*)\\[([' + nucleotides + '])>([' + extended_nucleotides + '])\\]([' + extended_nucleotides + ']*)',
        custom_motif.upper())
    if m:
        g = m.groups('')
        # print("GROUPS", m.group(1), m.group(2), m.group(3), m.group(4))
        entry = {}
        entry['logo'] = m.group(0)
        entry['motif'] = g[0] + g[1] + g[3]
        entry['position'] = len(g[0])
        entry['ref'] = g[1]
        entry['alt'] = g[2]
        if entry['ref'] == entry['alt']:
            return []
        entry['name'] = 'Custom motif'
        entry['references'] = ''
        return [entry, ]
    return []


def calculate_RR(ct):
    """
    Mutation is treatment
    No mutation is placebo

    :param ct: mutually exclusive counts of mutated matching motifs, matching mutations, matching motifs, and matching bases
    :return: enrichment or risk ratio
    """
    try:
        RR = ((ct.loc['mutation', 'motif'] / (ct.loc['mutation', 'motif'] + ct.loc['mutation', 'no motif'])) /
              (ct.loc['no mutation', 'motif'] / (ct.loc['no mutation', 'motif'] + ct.loc['no mutation', 'no motif'])))
    except ZeroDivisionError:
        RR = 0.0
    return RR


def calculate_RR_for_motif(ct):
    """
    Motif is treatment
    No motif is placebo

    :param ct: mutually exclusive counts of mutated matching motifs, matching mutations, matching motifs, and matching bases
    :return: enrichment or risk ratio
    """
    try:
        RR = ((ct.loc['mutation', 'motif'] / (ct.loc['mutation', 'motif'] + ct.loc['no mutation', 'motif'])) /
              (ct.loc['mutation', 'no motif'] / (ct.loc['mutation', 'no motif'] + ct.loc['no mutation', 'no motif'])))
    except ZeroDivisionError:
        RR = 0.0
    return RR


def calculate_OR(ct):
    """
    :param ct: mutually exclusive counts of mutated matching motifs, matching mutations, matching motifs, and matching bases
    :return: odds ratio
    """
    try:
        OR = (
            (ct.loc['mutation', 'motif'] / ct.loc['mutation', 'no motif']) /
            (ct.loc['no mutation', 'motif'] / ct.loc['no mutation', 'no motif']))
    except ZeroDivisionError:
        OR = 0.0
    return OR


def Haldane_correction(ct, pseudocount=0.5):
    """
    :param ct: mutually exclusive counts of mutated matching motifs, matching mutations, matching motifs, and matching bases
    :return: contigency table after Haldane correction is applied
    """
    """    apply Haldane correction (+ 0.5) if any of the values in the contingency table is zero """

    return ct + pseudocount if np.any(np.isclose(ct.to_numpy(), 0.0)) else ct


def calculate_mutation_load(N_mutations, enrichment):
    """
    Mutation load (minimum estimate) calculation following Gordenin et al protocol
    However, at this point motif matches are not filtered for p-value significance
    That's done in the end after multiple testing correction
    """
    mutation_load = 0.0
    if enrichment > 1.0:
        mutation_load = N_mutations * (enrichment - 1) / enrichment
    # elif p_value < p_value_threshold:   tests for enrichment depletion
    return mutation_load


def get_stats(ct, stat_type='fisher'):
    """
    Calculate Fisher and Chi2 test pvalues,
    :param ct: counts of mutated matching motifs, matching mutations, matching motifs, and matching bases
    :param stat_type: Type of pvalue (Fisher's ('fisher') or Chi-Square ('chi2'))
    :return: pvalue of the corresponding statistical test
    """
    p_val = 1.0

    if stat_type is None:
        stat_type = 'fisher'

    stat_type = stat_type.lower()

    acceptable_tests = ('fisher', 'chi2')
    if stat_type not in acceptable_tests:
        logger.warning('get_stats() can only calculate p-values for ' + str(acceptable_tests))

    if stat_type == 'fisher':
        try:
            p_val = stats.fisher_exact(ct, alternative="greater")[1]
            # if p_val > 0.05:
            #     p_val = stats.fisher_exact(ct, alternative="less")[1] #calculates if motif is underrepresented
        except ValueError:
            p_val = 1.0
    elif stat_type == 'chi2':
        try:
            p_val = stats.chi2_contingency(ct)[1]
        except ValueError:
            p_val = 1.0
    return p_val


def get_corrected_pvalues(p_values):
    qvalues = []
    if len(p_values):
        qvalues = multipletests(pvals=p_values, method='fdr_bh')[1]
    return qvalues


# @functools.lru_cache(maxsize=None)
def get_rev_comp_seq(sequence):
    """
    :param sequence: forward DNA sequence
    :return: reverse complimentary DNA sequence
    """
    # rev_comp_seq = "".join([complementary_nucleotide[i] for i in reversed(sequence)])
    cn = complementary_nucleotide
    return [(i[0], i[1], cn[i[2]], '-') for i in reversed(sequence)]


def mutated_base(mutation, ref, alt):
    """
    :param mutation: [(record.CHROM, record.POS, record.REF, record.ALT)]
    :param ref: list the nucleotide base pre-mutation
    :param alt: list the nucleotide base post-mutation
    :return: True if mutation matches the specified ref and alt
    """
    # makes sure single base substitution
    _, _, mut_ref, mut_alt = mutation
    if mut_alt and mut_ref and len(mut_ref) == 1 and len(mut_alt) == 1 and mut_ref != mut_alt:
        # mutation matches the substitution
        if mutation[2] in bases_dict[ref] and mutation[3] in bases_dict[alt]:
            return True


def find_matching_motifs(seq, motif, motif_position):
    """
    :param seq: DNA sequence
    :param motif: specified motif
    :param motif_position: position of mutated base in motif, 0-base numbering
    :return: generator of matching positions

    TODO: SLOW algorithm O(n * m). Need to create a suffix tree with regexp
    """
    # print("Looking for motif {} in {}, {}".format(motif, sequence, len(sequence) - len(motif)))
    for i in range(len(seq) - len(motif) + 1):
        # s = seq[i: i + len(motif)]
        # print(s)
        for j, c in enumerate(motif):
            if seq[i + j][2] not in bases_dict[c]:
                break
        else:
            yield seq[i + motif_position]


def find_matching_bases(seq, ref, motif, motif_position):
    """
    :param seq:
    :param ref:
    :param motif:
    :param motif_position:
    :return: bases that match mutations
    """
    for i in range(motif_position, len(seq) - (len(motif) - motif_position) + 1):
        # range excludes border of sequence that may be motifs that don't fit window size
        if seq[i][2] in bases_dict[ref]:
            yield seq[i]


def make_contingency_table(
        array=None,
        motif_mutation=None,
        no_motif_mutation=None,
        motif_no_mutation=None,
        no_motif_no_mutation=None):
    """ Make a 2x2 contingency table out of a numpy array or four integers"""

    if array is not None:
        assert isinstance(array, np.ndarray)
        assert array.shape == (2, 2)
    else:
        array = np.array([
            [motif_mutation, no_motif_mutation],
            [motif_no_mutation, no_motif_no_mutation]
        ])
    contingency_table = pd.DataFrame(array)
    contingency_table.columns = ["motif", "no motif"]
    contingency_table.index = ["mutation", "no mutation"]
    return contingency_table


def process_mutations(mutations, motif, motif_position, ref, alt, range_size, strand, stat_type=None):
    """
    :param mutations: mutations to be analyzed
    :param motif: specified motif to search for
    :param motif_position: location of mutation in motif, 0-base numbering from left of motif
    :param ref: base pre-mutation
    :param alt: base post-mutation
    :param range_size: how far in the motif to search for
    :param strand: strand motif should be searched on
    :param stat_type: type of pvalue: Fisher's (default) or Chi-Square
    :param dump_matches: an optional file handle to save all mutations matching the motif regardless of their significance
    :return: (results summary disctionary, data_dump with stored_data or None if dump_matches is None)
    """
    assert range_size >= 0
    assert len(ref) == 1
    assert len(alt) == 1
    assert 0 <= motif_position < len(motif)
    assert len(set(strand) - set("ATN")) == 0, "[process_mutations] only A, T, N allowed in strand parameter"

    matching_bases = set()
    matching_motifs = set()
    matching_mutated_motifs = set()
    matching_mutated_bases = set()

    # extra loop for sample in sample list
    for chrom, pos, transcript_strand, x, y, seq in mutations:
        # extract the longest sequence we would ever need (motif + range_size); range size = # bases outside mutation
        mutation = chrom, pos, x, y
        rev_seq = get_rev_comp_seq(seq)

        # assuming that all mutations are reported in '+' reference strand
        if strand == 'A' or (strand == 'T' and transcript_strand == '+') or (strand == 'N' and transcript_strand == '-'):
            # not mutated:
            for ref_match in find_matching_bases(seq, ref, motif, motif_position):
                matching_bases.add(ref_match[0:2])
            for motif_match in find_matching_motifs(seq, motif, motif_position):
                matching_motifs.add(motif_match[0:2])

            # mutated:
            if mutated_base(mutation, ref, alt):
                # m = (mutation[0], mutation[1], mutation[2], "+")
                matching_mutated_bases.add(mutation[0:2])

                context_of_mutation = seq[range_size - motif_position: range_size - motif_position + len(motif)]
                for motif_match in find_matching_motifs(context_of_mutation, motif, motif_position):
                    matching_mutated_motifs.add(motif_match[0:2])

        if strand == 'A' or (strand == 'T' and transcript_strand == '-') or (strand == 'N' and transcript_strand == '+'):
            # rev compl: not mutated:
            for ref_match in find_matching_bases(rev_seq, ref, motif, motif_position):
                matching_bases.add(ref_match[0:2])

            for motif_match in find_matching_motifs(rev_seq, motif, motif_position):
                matching_motifs.add(motif_match[0:2])

            # rev compl: mutated:
            if mutated_base(mutation, complementary_extended_nucleotide[ref], complementary_extended_nucleotide[alt]):
                # m = (mutation[0], mutation[1], mutation[2], "-")
                matching_mutated_bases.add(mutation[0:2])

                # rev comp:
                context_of_mutation = rev_seq[range_size - motif_position: range_size - motif_position + len(motif)]
                for motif_match in find_matching_motifs(context_of_mutation, motif, motif_position):
                    matching_mutated_motifs.add(motif_match[0:2])

    motif_mutation_count = len(matching_mutated_motifs)  # bases mutated in motif
    stat_mutation_count = len(matching_mutated_bases - matching_mutated_motifs)  # bases mutated not in motif
    stat_motif_count = len(matching_motifs - matching_mutated_motifs)  # bases not mutated in motif
    stat_ref_count = len(matching_bases - (matching_motifs | matching_mutated_bases))  # bases not mutated not in motif

    # number of A[T>G]T occurrences               motif_mutation_count
    # / number of [T>G] occurrences               stat_mutation_count + motif_mutation_count
    # ----------
    # number of ATT occurrences in DNA context    stat_motif_count
    # / number of T occurrences in DNA context    stat_ref_count + stat_motif_count

    contingency_table = make_contingency_table(
        motif_mutation=motif_mutation_count,
        no_motif_mutation=stat_mutation_count,
        motif_no_mutation=stat_motif_count,
        no_motif_no_mutation=stat_ref_count)

    # data={
    # "'{}>{}' mutation".format(ref, alt): [stat_mutation_count, motif_mutation_count],
    # "no '{}>{}' mutation".format(ref, alt): [stat_ref_count, stat_motif_count]},
    # index=("no '{}' motif".format(motif), "'{}' motif".format(motif)))
    logger.debug("\n" + contingency_table.to_string() + "\n")

    logger.debug("({} / ({} + {})  ) /  ({} / ({} + {}))".format(
        contingency_table.loc['mutation', 'motif'],
        contingency_table.loc['mutation', 'motif'],
        contingency_table.loc['mutation', 'no motif'],
        contingency_table.loc['no mutation', 'motif'],
        contingency_table.loc['no mutation', 'motif'],
        contingency_table.loc['no mutation', 'no motif']))

    contingency_table = Haldane_correction(contingency_table)

    enrichment = risk_ratio = calculate_RR(contingency_table)  # enrichment = risk ratio
    odds_ratio = calculate_OR(contingency_table)

    p_val = get_stats(contingency_table, stat_type)

    mut_load = calculate_mutation_load(motif_mutation_count, enrichment)

    result = {
        'enrichment': enrichment,  # AKA risk ratio
        'odds_ratio': odds_ratio,
        'mutation_load': math.ceil(mut_load),
        'pvalue': p_val,
        'bases_mutated_in_motif': motif_mutation_count,
        'bases_mutated_not_in_motif': stat_mutation_count,
        'bases_not_mutated_in_motif': stat_motif_count,
        'bases_not_mutated_not_in_motif': stat_ref_count,
        'total_mutations': len(mutations)
    }

    saved_matches = {
        'mutation_motif': matching_mutated_motifs
    }
    return result, saved_matches



import argparse
import sys
import logging
from mutagene.profiles.profile import calc_profile


logger = logging.getLogger(__name__)
genome_error_message = """requires genome name argument -g hg19, hg38, mm10, see http://hgdownload.cse.ucsc.edu/downloads.html for more
                          Use mutagene fetch to download genome assemblies"""


class ProfileMenu(object):
    def __init__(self, parser):
        parser.add_argument("--infile", "-i", nargs='*', help="Input file format", type=argparse.FileType('r'))
        parser.add_argument('--outfile', "-o", nargs='?', type=argparse.FileType('w'), default=sys.stdout,
                            help="Name of output file, will be generated in TSV format")
        parser.add_argument('--genome', "-g", help="Location of genome assembly file", type=str)
        parser.add_argument('--input-format', "-f", help="Input format: auto, MAF, VCF", type=str, default='auto')

        # for backwards compatibility with 0.8.X add a hidden action that would just take anything as a valid input
        parser.add_argument('action', nargs="?", metavar="")

    def callback(self, args):
        # print('ProfileMenu', args.action)
        self.calculate(args)

    def calculate(self, args):
        # print("Calculating...")
        if not args.infile:
            logger.warning("Provide input file in VCF or MAF format (-i) and a corresponding genome assembly (-g)")
            return
        if not args.genome:
            logger.warning(genome_error_message)
            return
        calc_profile(args.infile, args.outfile, args.genome, args.input_format)



import argparse
import sys
import logging

from mutagene.io.context_window import read_MAF_with_context_window, read_VCF_with_context_window
from mutagene.motifs import identify_motifs
from mutagene.io.motifs import write_motif_matches, get_known_motifs


logger = logging.getLogger(__name__)
genome_error_message = """requires genome name argument -g hg19, hg38, mm10, see http://hgdownload.cse.ucsc.edu/downloads.html for more
                        Use mutagene fetch to download genome assemblies"""


class MotifMenu(object):
    def __init__(self, parser):
        parser.description = ""
        parser.epilog = """
Examples:
# search in sample2.vcf for all preidentified motifs in mutagene using hg19
mutagene motif --infile sample2.vcf --input-format VCF --genome hg19

# search for the presence of the C[A>T] motif in sample1.maf using hg19 not checking for strand-specificity
mutagene motif --infile sample1.maf --input-format MAF --genome hg19 --motif 'C[A>T]' --strand A
        """

        ###################################################################
        required_group = parser.add_argument_group('Required arguments')
        required_group.add_argument(
            "--infile", "-i", help="Input file in MAF or VCF format with one or multiple samples",
            type=argparse.FileType('r'))
        required_group.add_argument(
            '--genome', "-g", help="Location of genome assembly file in 2bit format", type=str)

        ###################################################################
        optional_group = parser.add_argument_group('Optional arguments')
        optional_group.add_argument('--input-format', "-f", help="Input format: MAF, VCF", type=str, choices=['MAF', 'VCF'], default='MAF')
        optional_group.add_argument(
            "--motif", "-m",
            help="Motif to search for, use the 'R[C>T]GY' syntax for the motif. Use quotes", type=str)
        optional_group.add_argument(
            '--outfile', "-o", nargs='?', type=argparse.FileType('w'), default=sys.stdout,
            help="Name of output file, will be generated in TSV format")

        # for backwards compatibility with 0.8.X add a hidden action that would just take anything as a valid input
        optional_group.add_argument('action', nargs="?", metavar="")

        ###################################################################
        advanced_group = parser.add_argument_group('Advanced arguments')
        advanced_group.add_argument(
            '--window-size', "-w", help="Context window size for motif search, default setting is 50",
            type=int, default=50)
        advanced_group.add_argument(
            '--strand', "-s",
            help="Transcribed strand (T), non-transcribed (N), any (A), or all (TNA default) ",
            type=str, default='TNA', choices=['T', 'N', 'A', 'TNA'])
        advanced_group.add_argument(
            '--threshold', "-t",
            help="Significance threshold for qvalues, default value=0.05",
            type=float, default=0.05)
        advanced_group.add_argument(
            '--save-motif-matches',
            help="Save mutations in matching motifs to a BED file",
            type=argparse.FileType('w'), default=None)
        advanced_group.add_argument(
            '--test',
            help="Statistical test to use",
            type=str, default='Fisher', choices=['Fisher', 'Chi2'])

        self.parser = parser

    @classmethod
    def search(cls, args):
        if not args.infile:
            logger.warning("Provide input file in VCF or MAF format (-i) and a corresponding genome assembly (-g)")
            return
        if not args.genome:
            logger.warning(genome_error_message)
            return
        if args.threshold > 1.0 or args.threshold < 0.0:
            logger.warning("The threshold value should be between 0.0 and 1.0, inclusive")
            return
        if not args.motif:
            logger.info("Searching for predefined motifs")
            logger.info("Will reduce statistical power because of multiple comparisons")
            custom_motif = None
        else:
            custom_motif = args.motif
            custom_motif = custom_motif.replace('-', '>')
            custom_motif = custom_motif.replace('/', '>')
            custom_motif = custom_motif.replace('.', '>')
            custom_motif = custom_motif.replace('->', '>')

            if ">" not in custom_motif or "]" not in custom_motif:
                logger.warning(
                    "Mutagene motif search failed because motif cannot be processed."
                    "Check to make sure motif input is in quotes")
                return

            logger.info("Searching for motif {}".format(custom_motif))

        if args.window_size > 5000 or args.window_size < 1:
            logger.warning('window-size should be between 1 and 5000 nucleotides')
            return

        try:
            if args.input_format == 'VCF':
                mutations, mutations_with_context, processing_stats = read_VCF_with_context_window(
                    args.infile, args.genome, args.window_size)
            elif args.input_format == 'MAF':
                mutations, mutations_with_context, processing_stats = read_MAF_with_context_window(
                    args.infile, args.genome, args.window_size)
        except ValueError as e:
            logger.warning('Not able to parse input file in {} format: {}. You can specify a different format with --input-format (-f)'.format(args.input_format, e))
            sys.exit(1)

        if len(mutations_with_context) == 0:
            logger.warning("No mutations loaded")

        #######################
        # Performance PROFILING
        # import cProfile
        # import pstats
        # pr = cProfile.Profile()
        # pr.enable()

        matching_motifs = identify_motifs(
            samples_mutations=mutations_with_context,
            custom_motif=custom_motif,
            strand=args.strand,
            threshold=args.threshold,
            dump_matches=args.save_motif_matches,
            stat_type=args.test) if mutations_with_context is not None else []

        #######################
        # Performance PROFILING
        # pr.disable()
        # p = pstats.Stats(pr)
        # p.sort_stats('ncalls').print_stats(20)  # strip_dirs()

        if len(matching_motifs) == 0:
            logger.warning("No significant motif matches found")
        else:
            write_motif_matches(args.outfile, matching_motifs)

    def list(self, args):
        """ Prints the list of known motifs bundled with the package"""
        print("\nThe list of mutational motifs that will be tested by default:")
        for m in get_known_motifs():
            print("{:20}\t{}".format(m['name'], m['logo']))
        print("any custom motif can be specified with --motif (-m)")

    def callback(self, args):
        # getattr(cls, args.action)(args)
        if args.infile:
            MotifMenu.search(args)
        else:
            self.parser.print_usage()
            self.list(args)
            sys.exit(1)



# from tqdm import tqdm
from collections import defaultdict
from mutagene.dna import nucleotides
from mutagene.dna import complementary_trinucleotide, complementary_nucleotide

import os
import numpy as np

import logging
logger = logging.getLogger(__name__)


def read_profile_file(profile_file):
    """Read profile from file

    Format: T[A>C]G frequency

    Arguments:
        profile_file {str} -- profile file name

    Returns:
        mutations, stats --
    """
    try:
        with open(profile_file) as f:
            profile_str = f.read()
            return read_profile_str(profile_str)
    except IOError:
        logger.warning("Could not read profile file " + profile_file)
        return None


def read_profile_str(profile_str):
    mutations = defaultdict(float)
    for line in profile_str.splitlines():
        if len(line) == 0:
            continue
        if line.startswith('#'):
            continue
        fields = line.strip().upper().split()
        if len(fields) != 2:
            return None, None

        if len(fields[0]) != 7:
            return None, None

        if fields[0][1] != "[" or fields[0][3] != ">" or fields[0][5] != "]":
            return None, None

        p5, _, x, _, y, _, p3 = tuple(fields[0])

        if p5 not in nucleotides or p3 not in nucleotides or y not in nucleotides:
            return None, None

        if x not in "TC":
            return None, None

        try:
            f = float(fields[1])
        except:
            return None, None

        mutations[p5 + p3 + x + y] = f

    values = []
    for p5 in nucleotides:
        for p3 in nucleotides:
            for x in "CT":
                for y in nucleotides:
                    if x != y:
                        values.append(mutations.get(p5 + p3 + x + y, 0.0))
    return values


"""
Type,SubType,SBS1,SBS2,SBS3,SBS4,SBS5,SBS6,SBS7a,SBS7b,SBS7c,SBS7d,SBS8,SBS9,SBS10a,SBS10b,SBS11,SBS12,SBS13,SBS14,SBS15,SBS16,SBS17a,SBS17b,SBS18,SBS19,SBS20,SBS21,SBS22,SBS23,SBS24,SBS25,SBS26,SBS27,SBS28,SBS29,SBS30,SBS31,SBS32,SBS33,SBS34,SBS35,SBS36,SBS37,SBS38,SBS39,SBS40,SBS41,SBS42,SBS43,SBS44,SBS45,SBS46,SBS47,SBS48,SBS49,SBS50,SBS51,SBS52,SBS53,SBS54,SBS55,SBS56,SBS57,SBS58,SBS59,SBS60,SBS84,SBS85
C>A,ACA,8.86E-04,5.80E-07,2.08E-02,4.22E-02,1.20E-02,4.25E-04,6.70E-05,2.33E-03,4.83E-03,4.04E-05,4.41E-02,5.58E-04,2.19E-03,1.82E-04,1.46E-04,4.52E-03,1.82E-03,1.12E-03,9.44E-04,1.60E-02,2.07E-03,6.08E-04,5.15E-02,1.27E-03,6.19E-04,1.57E-04,6.01E-03,8.35E-04,3.64E-02,9.90E-03,8.73E-04,5.21E-03,7.84E-04,6.32E-02,1.80E-03,9.54E-03,2.23E-02,3.11E-03,4.87E-03,8.83E-03,2.52E-02,3.95E-03,1.28E-02,1.17E-02,2.82E-02,2.11E-03,1.16E-03,2.95E-02,7.68E-18,9.11E-03,4.40E-03,6.78E-02,8.55E-04,2.51E-02,1.19E-01,1.41E-01,1.52E-02,5.38E-03,2.16E-03,5.88E-03,1.26E-02,1.23E-02,5.89E-02,3.59E-03,6.15E-03,0.003471994832,0.006080257390
"""
"""
Sequencing artifacts:

Possible sequencing artefacts
"""
def read_COSMICv3_signatures():
    sequencing_artifacts = [
        "SBS27", "SBS43", "SBS45", "SBS46", "SBS47", "SBS48", "SBS49", "SBS50", "SBS51",
        "SBS52", "SBS53", "SBS54", "SBS55", "SBS56", "SBS57", "SBS58", "SBS59", "SBS60"]

    mutations = defaultdict(dict)
    dirname = os.path.dirname(os.path.realpath(__file__))
    fname = dirname + "/../data/signatures/sigProfiler_SBS_signatures_2019_05_22.csv"
    signature_names = []
    with open(fname, 'r') as f:
        for i, line in enumerate(f):
            fields = line.strip().split(',')
            if i == 0:
                signature_names = fields[2:]
                continue

            x, y = fields[0].split(">")
            p5, _, p3 = tuple(fields[1])

            for j, val in enumerate(fields[2:]):
                val = float(val)
                mutations[j][p5 + p3 + x + y] = val

    W = []
    for j in range(len(signature_names)):
        profile = []
        for p5 in nucleotides:
            for p3 in nucleotides:
                for x in "CT":
                    for y in nucleotides:
                        if x != y:
                            profile.append(mutations[j].get(p5 + p3 + x + y, 0.0))
        W.append(profile)
    W = np.array(W).T
    return W, signature_names


def read_KUCAB_signatures():
    """
    A compendium of mutational signatures of environmental agents Kucab et al.  Serena Nik-Zainal
    DOI: 10.17632/m7r4msjb4c.2
    https://data.mendeley.com/datasets/m7r4msjb4c/2#folder-9f2942fa-935b-4ac3-aa26-35784fb68a85

    MutationType    Potassium bromate (875 uM)  DBADE (0.109 uM)    Formaldehyde (120 uM)   Semustine (150 uM)  Temozolomide (200 uM)   DMH (11.6 mM) + S9  Benzidine (200 uM)  DBP (0.0039 uM) MX (7 uM) + S9  Methyleugenol (1.25 mM) 4-ABP (300 uM) + S9 DBPDE (0.000156 uM) DBP (0.0313 uM) + S9    DBADE (0.0313 uM)   1,8-DNP (0.125 uM)  BPDE (0.125 uM) MNU (350 uM)    ENU (400 uM)    Cyclophosphamide (18.75 uM) + S9    BaP (0.39 uM) + S9  6-Nitrochrysene (12.5 uM) + S9  AAI (1.25 uM)   Potassium bromate (260 uM)  6-Nitrochrysene (0.78 uM)   Ellipticine (0.375 uM) + S9 DBA (75 uM) + S9    PhIP (3 uM) + S9    AFB1 (0.25 uM) + S9 3-NBA (0.025 uM)    1,6-DNP (0.09 uM)   5-Methylchrysene (1.6 uM) + S9  Furan (100 mM) + S9 SSR (1.25 J)    AAII (37.5 uM)  Propylene oxide (10 mM) N-Nitrosopyrrolidine (50 mM)    Mechlorethamine (0.3 uM)    DES (0.938 mM)  DMS (0.078 mM)  Cisplatin (3.125 uM)    OTA (0.08 uM) + S9  Carboplatin (5 uM)  DBAC (5 uM) + S9    Temozolomide (200 uM).1 Cisplatin (12.5 uM) AZD7762 (1.625 uM)  3-NBA (0.1 uM)  PhIP (4 uM) + S9    BaP (2 uM) + S9 6-Nitrochrysene (50 uM) + S9    6-Nitrochrysene (50 uM) 1,8-DNP (8 uM)  DBPDE (0.000625 uM) Control
    A[C>A]A 0.109398510425891   0.0240330588511261  0.0181865586268237  0.000122421190463747    0   0.00665025743214769 0.0564462237667514  0   0.0494977586860332  0.111458492403729   0.00483165426221388 0.0002218019424568  0.00869925685873544 0.0264652966450285  0.0195599544767651  0.0237017339519746  0.00262256704780487 0.0130480902574389  0.00555898815479868 0.0406568174086766  0.0100871864208784  0.00428520931740606 0.114738757387778   0.00515014598157648 0.0482801050571615  0.0323606741641645  0.0279971115871403  0.0379396506608179  0.000340069771278152    0.000751844456779883    0.0286278352872072  0.0503953181060661  2.22941465639634e-05    0.000162365803469566    0.0104143358776828  0.00685630323850839 0.000133921576621699    0.014166642071802   0.000461507809128427    0.00279492689318811 0.11584205195812    0.00194170922837543 0.00673772703708239 0.000256609057933161    1.12145020078519e-05    0.000395123818576932    0.0258298926507019  0.032667880519281   0.0283815533748292  0.00855231498108648 0.00555769823738427 0.0180349250334742  0.0186419336589614  0.0605073306958345
    """
    mutations = defaultdict(dict)
    dirname = os.path.dirname(os.path.realpath(__file__))
    fname = os.path.normpath(dirname + "/../data/signatures/Kucab.txt")
    signature_names = []
    with open(fname, 'r') as f:
        for i, line in enumerate(f):
            fields = line.strip().split("\t")
            if i == 0:
                signature_names = fields[1:]
                continue

            # parse A[C>A]A
            m = fields[0]
            p5 = m[0]
            assert m[1] == '['
            x = m[2]
            assert m[3] == '>'
            y = m[4]
            assert m[5] == ']'
            p3 = m[6]

            for j, val in enumerate(fields[1:]):
                val = float(val)
                mutations[j][p5 + p3 + x + y] = val

    W = []
    for j in range(len(signature_names)):
        profile = []
        for p5 in nucleotides:
            for p3 in nucleotides:
                for x in "CT":
                    for y in nucleotides:
                        if x != y:
                            profile.append(mutations[j].get(p5 + p3 + x + y, 0.0))
        W.append(profile)
    W = np.array(W).T
    return W, signature_names


def _read_mutagene_signatures(name, n_signatures):
    """ Read signatures preprocessed for mutagene website / package """
    dirname = os.path.dirname(os.path.realpath(__file__))
    W = []
    signature_names = []
    for i in range(n_signatures):
        fname = dirname + "/../data/signatures/{}_{}.profile".format(name, i + 1)
        profile = read_profile_file(fname)
        W.append(profile)
        signature_names.append("{}".format(i + 1))

    W = np.array(W).T
    return W, signature_names


def read_MGA_signatures():
    return _read_mutagene_signatures('A', 5)


def read_MGB_signatures():
    return _read_mutagene_signatures('B', 10)


def read_COSMICv2_signatures():
    return _read_mutagene_signatures('C', 30)


def read_signatures(name, only=None):
    """
    Retrieve a set of signatures by its name or number of signatures specified as 'str'.
    Returns tuple: numpy matrix and list of names
    """

    # number of sigatures matching to names
    signatures_dict = {
        '5': 'MGA',
        '10': 'MGB',
        '30': 'COSMICv2',
        '49': 'COSMICv3',
        '53': 'KUKAB'
    }
    inv_signatures_dict = dict(zip(signatures_dict.values(), signatures_dict.keys()))

    if name in signatures_dict:
        name = signatures_dict[name]

    assert name in inv_signatures_dict, "Unknown name for a signature set: {}".format(name)

    function_name = "read_{}_signatures".format(name)
    W, signature_names = globals()[function_name]()

    # filter by list of 'only' signatures,
    # append signature set prefix to signature names
    delete_signatures = []
    filtered_signature_names = []
    for i, x in enumerate(signature_names):
        if only is not None:
            if x not in only:
                delete_signatures.append(i)
                continue
        filtered_signature_names.append("{}-{}".format(name, x))

    if only is not None:
        W = np.delete(W, np.array(delete_signatures), axis=1)

    return W, filtered_signature_names


def write_profile(profile_file, p, counts=True):
    with open(profile_file, 'w') as o:
        write_profile_file(o, p, counts)


def format_profile(values, counts=False):
    if counts:
        format_str = "{}[{}>{}]{}\t{:d}\n"
    else:
        format_str = "{}[{}>{}]{}\t{:.12f}\n"
    attrib = get_profile_attributes_dict()
    result = ""
    for i, v in enumerate(values):
        x, y = attrib[i]['mutation']
        p5, p3 = attrib[i]['context']
        if counts:
            v = int(v)
        result += format_str.format(p5, x, y, p3, v)
    return result


def write_profile_file(file_handle, p, counts=True):
    formatted_profile = format_profile(p, counts)
    file_handle.write(formatted_profile)


def get_profile_attributes_dict(signature_order=False):
    attribs = []
    if signature_order:
        for x in "CT":
            for y in nucleotides:
                if x != y:
                    for p5 in nucleotides:
                        for p3 in nucleotides:
                            attribs.append({
                                'mutation': x + y,
                                'context': p5 + p3
                            })
    else:
        for p5 in nucleotides:
            for p3 in nucleotides:
                for x in "CT":
                    for y in nucleotides:
                        if x != y:
                            attribs.append({
                                'mutation': x + y,
                                'context': p5 + p3
                            })
    return attribs


def get_attributes():
    attribs = get_profile_attributes_dict()
    for i, v in enumerate(attribs):
        x, y = v['mutation']
        p5, p3 = v['context']
        attribs[i]['mut'] = "{}{}".format(x, y),
        attribs[i]['mutation'] = "{}{}{} → {}{}{}".format(p5, x, p3, p5, y, p3),
        attribs[i]['context'] = "{}{}".format(p5, p3)
    return attribs


def profile_to_dict(counts_profile):
    mutation_prob = defaultdict(dict)
    total_count = sum(counts_profile)
    freqs = [x / total_count for x in counts_profile]

    for i, attrib in enumerate(get_profile_attributes_dict()):
        x, y = attrib['mutation']
        p5, p3 = attrib['context']
        tri = p5 + x + p3
        mutation_prob[tri][y] = freqs[i]
        mutation_prob[complementary_trinucleotide[tri]][complementary_nucleotide[y]] = freqs[i]
    return mutation_prob



from itertools import product

nucleotides = "ACGT"  # this order of nucleotides is important for reversing
mutation_contexts = [a + b for a in nucleotides for b in nucleotides]
amino_acids = "ACDEFGHIKLMNPQRSTVWY"
amino_acids_with_stop = amino_acids + "*"
# complementary_nucleotides = reversed(nucleotides)
complementary_nucleotide = dict(zip(nucleotides, reversed(nucleotides)))
complementary_nucleotide['N'] = 'N'

complementary_context = {x: complementary_nucleotide[x[1]] + complementary_nucleotide[x[0]] for x in mutation_contexts}
complementary_trinucleotide = dict(zip(
    ["".join(x) for x in product(nucleotides, repeat=3)],
    ["".join(reversed(x)) for x in product(reversed(nucleotides), repeat=3)]))

bases_dict = {
    "A": "A", "G": "G", "T": "T", "C": "C",
    "W": "AT", "S": "CG", "M": "AC", "K": "GT", "R": "AG", "Y": "CT",
    "B": "TCG", "D": "AGT", "H": "ACT", "V": "ACG", "N": "ATGC"}

extended_nucleotides = "ACTGWSMKRYBDHVN"
complementary_extended_nucleotide = dict(zip(extended_nucleotides, "TGACWSKMYRVHDBN"))

comp_dict = {
    "A": "T", "T": "A", "C": "G", "G": "C",
    "W": "AT", "S": "CG", "K": "AC", "M": "GT", "Y": "AG", "R": "CT",
    "V": "TCG", "H": "AGT", "D": "ACT", "B": "ACG", "N": "ATGC"}

codon_table = {
    "GCT": "A", "GCC": "A", "GCA": "A", "GCG": "A",
    "TTA": "L", "TTG": "L", "CTT": "L", "CTC": "L", "CTA": "L", "CTG": "L",
    "CGT": "R", "CGC": "R", "CGA": "R", "CGG": "R", "AGA": "R", "AGG": "R",
    "AAA": "K", "AAG": "K",
    "AAT": "N", "AAC": "N",
    "ATG": "M",
    "GAT": "D", "GAC": "D",
    "TTT": "F", "TTC": "F",
    "TGT": "C", "TGC": "C",
    "CCT": "P", "CCC": "P", "CCA": "P", "CCG": "P",
    "CAA": "Q", "CAG": "Q",
    "TCT": "S", "TCC": "S", "TCA": "S", "TCG": "S", "AGT": "S", "AGC": "S",
    "GAA": "E", "GAG": "E",
    "ACT": "T", "ACC": "T", "ACA": "T", "ACG": "T",
    "GGT": "G", "GGC": "G", "GGA": "G", "GGG": "G",
    "TGG": "W",
    "CAT": "H", "CAC": "H",
    "TAT": "Y", "TAC": "Y",
    "ATT": "I", "ATC": "I", "ATA": "I",
    "GTT": "V", "GTC": "V", "GTA": "V", "GTG": "V",
    "TAG": "*", "TGA": "*", "TAA": "*"
}

exome_trinucleotides = {
    "GCA": 1870205, "ACT": 1317607, "GCC": 2018826, "CCT": 2045943, "GTC": 1272257, "ATT": 1226179, "CTC": 1989259,
    "ACA": 1614930, "ATC": 1343451, "ACG": 602918, "TTC": 1901940, "GTT": 1149725, "GCG": 857006, "GTG": 1685105,
    "ACC": 1369960, "CCA": 2359526, "TTG": 1588902, "ATA": 841828, "TCA": 1853413, "CCG": 1009679, "TTA": 774505,
    "TCG": 640487, "ATG": 1654761, "GTA": 798348, "CTT": 1881403, "GCT": 1983552, "CTA": 713831, "TTT": 1756413,
    "CCC": 1827705, "TCC": 2026380, "TCT": 2000322, "CTG": 2769315}

aa_short = amino_acids_with_stop
aa_long = ["Ala", "Leu", "Pro", "Gly", "Met", "Ser", "Thr", "Trp", "Ile", "Val", "Cys", "Asp", "Glu", "Phe", "His", "Lys", "Asn", "Gln", "Arg", "Tyr", "STOP"]
aa_dict = dict(zip(aa_short, aa_long))

bases_dict = {"A": "A", "G": "G", "T": "T", "C": "C", "W": "AT", "S": "CG", "M": "AC", "K": "GT", "R": "AG", "Y": "CT",
              "B": "TCG", "D": "AGT", "H": "ACT", "V": "ACG", "N": "ATGC"}

comp_dict = {"A": "T", "T": "A", "C": "G", "G": "C", "W": "AT", "S": "CG", "K": "AC", "M": "GT", "Y": "AG",
             "R": "CT", "V": "TCG", "H": "AGT", "D": "ACT", "B": "ACG", "N": "ATGC"}

chromosome_name_mapping = {
    "chr23": "chrX",
    "chr24": "chrY",
    "chr25": "chrXY",
    "chr26": "chrM",
}




from mutagene.dna import nucleotides


def get_dummy_signatures_lists():
    """
    Generate 6 dummy signatures
    Each will have uniform non-zero frequencies corresponding to one mutation type
    Format them as lists
    """
    dummy_signatures = []
    for mutation in (("C", "A"), ("C", "T"), ("C", "G"), ("T", "A"), ("T", "C"), ("T", "G")):
        values = []
        for p5 in nucleotides:
            for p3 in nucleotides:
                for x in "CT":
                    for y in nucleotides:
                        if x != y:
                            if mutation == (x, y):
                                values.append(1.0 / 16.0)
                            else:
                                values.append(0.0)
        name = mutation[0] + " to " + mutation[1]
        dummy_signatures.append((name, values))
    return dummy_signatures



#
# Code copied from mutagene website
#

# from sqlalchemy.sql.expression import bindparam
# from collections import defaultdict

# from .mutations import *
# from ..database.models import *
# from ..database import Session as db


def export_cohorts(PATH):

    ######################################################
    # Export profiles
    ######################################################
    profiles = []
    q = db.query(Signature.id, Signature.name, Signature.mut_type, Signature.n_mutations,
                 Signature.nr_mutations, Signature.n_samples, Signature.signature_type, Signature.cancer_type_id,
                 CancerType.name.label("cancer_type"), PrimarySite.name.label("primary_site"),
                 CancerType.code.label("cancer_code")).\
        join(CancerType).join(PrimarySite).\
        filter(Signature.signature_type == "CT").\
        filter(Signature.mut_type == "A").\
        filter(Signature.normalization == "MUT").\
        filter(Signature.active == 1).\
        order_by(Signature.name)
    profiles.extend(q)

    q = db.query(Signature.id, Signature.name, Signature.mut_type, Signature.n_mutations,
                 Signature.nr_mutations, Signature.n_samples, Signature.signature_type,
                 bindparam("cancer_type", "Pan-cancer"), bindparam("primary_site", "All tissues"),
                 bindparam("cancer_code", ""), bindparam("cancer_projects", "")).\
        filter(Signature.signature_type == "PC").\
        filter(Signature.mut_type == "A").\
        filter(Signature.normalization == "MUT").one()
    profiles.append(q)

    # Bening Pan-tissue
    q = db.query(Signature.id, Signature.name, Signature.mut_type, Signature.n_mutations,
                 Signature.nr_mutations, Signature.n_samples, Signature.signature_type,
                 bindparam("cancer_type", "Benign"), bindparam("primary_site", "All tissues"),
                 bindparam("cancer_code", ""), bindparam("cancer_projects", "")).\
        filter(Signature.signature_type == "BEN").\
        filter(Signature.mut_type == "A").\
        filter(Signature.primary_site == None).one()
    profiles.append(q)

    # dbSNP
    q = db.query(Signature.id, Signature.name, Signature.mut_type, Signature.n_mutations,
                 Signature.nr_mutations, Signature.n_samples, Signature.signature_type,
                 bindparam("cancer_type", "SNP"), bindparam("primary_site", "Germline"),
                 bindparam("cancer_code", ""), bindparam("cancer_projects", "")).\
        filter(Signature.signature_type == "SNP").\
        filter(Signature.mut_type == "A").\
        filter(Signature.primary_site == None).one()
    profiles.append(q)

    # profiles.extend(q)
    # profiles.extend(q)

    # for each cancer type
    all_cancer_types = []
    for profile in profiles:
        # skip the second LUAD dataset
        if profile.id == 34904:
            continue
        all_cancer_types.append(profile.id)
        print("Processing {}".format(profile.name))
        data = {}
        s = db.query(Signature).get(profile.id)
        for d in s.data:
            idx = d.c53 + d.XY
            # data[idx] = d.freq
            data[idx] = d.count

        result = ""
        for p5 in nucleotides:
            for p3 in nucleotides:
                for x in "CT":
                    for y in nucleotides:
                        if x != y:
                            result += "{}[{}>{}]{}\t{}\n".format(p5, x, y, p3, data.get(p5 + p3 + x + y, 0), 8)

        fname = "{}/{}.profile".format(PATH, profile.name.replace(" ", "_"))
        with open(fname, 'w') as o:
            o.write("#NSAMPLES\t{}\n".format(profile.n_samples))
            o.write(result)

        ######################################################
        # Export mutations for each profile
        ######################################################
        if profile.signature_type == "CT":
            mm = db.query(Mutation).join(SignatureMutation).filter(SignatureMutation.signature_id == int(profile.id))
        elif profile.signature_type == "PC":
            mm = db.query(Mutation).join(SignatureMutation).filter(SignatureMutation.signature_id.in_(all_cancer_types))
        aa_mutations = defaultdict(int)
        na_mutations = defaultdict(int)

        for m in mm:
            if len(m.grc_pos) == 0 or len(m.mut_aa) == 0:
                continue
            chrom, pos = m.grc_pos.split('-')[0].split(':')
            # print(m.mut_aa)
            # print(m.mut_cds)
            mut_aa = m.mut_aa.split('.')[1]
            X, Y = m.context.XY
            aa_mutations[(m.gene.name, mut_aa)] += 1
            na_mutations[(chrom, pos, X, Y)] += 1

        if len(aa_mutations):
            with open("{}/{}.aa_mutations.txt".format(PATH, profile.name.replace(" ", "_")), 'w') as aa:
                for key, val in aa_mutations.items():
                    aa.write("{}\t{}\t{}\n".format(*key, val))

        if len(na_mutations):
            with open("{}/{}.dna_mutations.txt".format(PATH, profile.name.replace(" ", "_")), 'w') as na:
                for key, val in na_mutations.items():
                    na.write("{}\t{}\t{}\t{}\t{}\n".format(*key, val))



import csv
from collections import namedtuple

import twobitreader as tbr

from tqdm import tqdm
from collections import defaultdict

from mutagene.dna import nucleotides, complementary_nucleotide

import logging
logger = logging.getLogger(__name__)


def get_context_53_twobit(mutations, twobit_file):
    """
    User twobitreader to get context of mutations
    """
    contexts = {}

    fname = twobit_file if twobit_file.endswith('.2bit') else twobit_file + '.2bit'
    f = tbr.TwoBitFile(fname)

    cn = complementary_nucleotide
    for (chrom, pos, x, y) in tqdm(mutations, leave=False):
        start = int(pos) - 1  # zero-based numbering
        chrom = str(chrom)
        chromosome = chrom if chrom.startswith('chr') else 'chr' + chrom

        nuc5 = 'N'
        nuc3 = 'N'
        nuc = 'N'
        if chromosome in f:
            try:
                seq = f[chromosome][start - 1: start + 2]  # +/- 1 nucleotide
                nuc5, nuc, nuc3 = tuple(seq.upper())
            except:
                nuc = 'N'
                # print(chromosome, x, nuc5, nuc, nuc3)
            if nuc != 'N' and nuc != x:
                if cn[nuc] == x:
                    nuc3 = cn[nuc5]
                    nuc5 = cn[nuc3]
                else:
                    nuc3 = nuc5 = 'N'
        else:
            # print("NO CHROM", chromosome)
            pass
        contexts[(chrom, pos)] = (nuc5, nuc3)
    return contexts


def get_context_batch(mutations, assembly, method='twobit'):
    """
        Get context for a list of mutations [(chrom, pos, x, y) ] format
    """
    if assembly is None:
        assembly = 38

    if method is None:
        method = 'twobit'

    methods = {
        # 'ensembl': get_context_ensembl,
        'twobit': get_context_53_twobit
    }

    contexts = methods[method](mutations, assembly)
    return contexts


def read_auto_profile(muts, fmt, asm):
    # if isinstance(muts, io.TextIOWrapper):
    #     print("OOOO")
    #     muts = muts.readlines()
    # print(muts)
    mutations = None
    processing_stats = None
    if fmt is not None:
        fmt = fmt.upper()

    if fmt is None or fmt == "AUTO" or fmt == 'auto' or fmt == "":
        mutations_lines = []
        for line in muts:
            mutations_lines.append(line.strip())
            if line.startswith("#version 2."):
                fmt = "MAF"
                break
            if len(line.strip()) == 0 or line.startswith("#"):
                continue
            tabs = line.split()
            if len(tabs) == 2 and "[" in tabs[0] and "]" in tabs[0]:
                fmt = "PROFILE"
                break
            if (len(tabs) == 3 and tabs[1] == ">") or (len(tabs) == 1 and ">" in tabs[0]):
                fmt = "TRI"
                break
            if len(tabs) > 3:
                fmt = "VCF"
                if tabs[0].lower().startswith("chr"):
                    break  # yes, it's VCF
                if tabs[4].lower().startswith("chr"):
                    fmt = "MAF"
                    break
        mutations_lines.extend(muts.readlines())
    else:
        mutations_lines = muts.readlines()

    logger.info("DATA FORMAT:" + fmt)

    if fmt not in ['MAF', 'VCF']:
        logger.warning("The dataformat [{}] is not supported".format(fmt))

    if fmt == "VCF":
        mutations, processing_stats = read_VCF_profile(mutations_lines, asm)
    if fmt == "MAF":
        mutations, processing_stats = read_MAF_profile(mutations_lines, asm)

    return mutations, processing_stats


def read_MAF_profile(muts, asm):

    cn = complementary_nucleotide
    mutations = defaultdict(float)
    N_skipped = 0

    try:
        reader = csv.reader((row for row in muts if not row.startswith('#')), delimiter='\t')
        # get names from column headers
        header = next(reader)
        # MAF = namedtuple("MAF", map(str.lower, next(reader)))
        header = tuple(map(lambda s: s.lower().replace('.', '_'), header))
        # print(header)
        MAF = namedtuple("MAF", header, rename=True)
    except ValueError:
        # raise
        logger.warning("MAF format not recognized")
        return mutations, {}

    N_loaded = N_skipped = 0

    raw_mutations = []
    for data in map(MAF._make, reader):
        # assembly_build = col_list[3]  # MAF ASSEMBLY
        # strand = col_list[7]    # MAF STRAND
        # ID = col_list[2]

        # chromosome is expected to be one or two number or one letter
        chrom = data.chromosome  # MAF CHROM
        if chrom.lower().startswith("chr"):
            chrom = chrom[3:]
        # if len(chrom) == 2 and chrom[1] not in "0123456789":
        #     chrom = chrom[0]

        pos = int(data.start_position)    # MAF POS START
        pos_end = int(data.end_position)  # MAF POS END
        x = data.reference_allele         # MAF REF

        y1 = data.tumor_seq_allele1       # MAF ALT1
        y2 = data.tumor_seq_allele2       # MAF ALT2

        if pos != pos_end:
            continue

        # skip if found unexpected nucleotide characters
        if len(set([x, y1, y2]) - set(nucleotides)) > 0:
            continue
        # print("OK")

        y = y1 if y1 != x else None
        y = y2 if y2 != x else y
        if y is None:
            continue

        raw_mutations.append((chrom, pos, x, y))

    # MAX = 100000000
    if len(raw_mutations) > 0:
        # if len(raw_mutations) > MAX:
        #     raw_mutations = raw_mutations[:MAX]

        contexts = get_context_batch(raw_mutations, asm)

        if contexts is None:
            return None, None

        if len(contexts) == 0:
            return None, None

        for (chrom, pos, x, y) in raw_mutations:
            p5, p3 = contexts.get((chrom, pos), ("N", "N"))

            if len(set([p5, x, y, p3]) - set(nucleotides)) > 0:
                # print("Skipping invalid nucleotides")
                N_skipped += 1
                continue

            if x in "CT":
                mutations[p5 + p3 + x + y] += 1.0
            else:
                # complementary mutation
                mutations[cn[p3] + cn[p5] + cn[x] + cn[y]] += 1.0

    N_loaded = int(sum(mutations.values()))
    processing_stats = {'loaded': N_loaded, 'skipped': N_skipped, 'format': 'MAF'}
    return mutations, processing_stats


def read_VCF_profile(muts, asm=None):
    cn = complementary_nucleotide
    mutations = defaultdict(float)
    N_skipped = 0

    raw_mutations = []
    for i, line in enumerate(muts):
        if line.startswith("#"):
            continue
        if len(line) < 10:
            continue

        col_list = line.split()
        if len(col_list) < 4:
            continue

        # ID = col_list[2]
        # chromosome is expected to be one or two number or one letter
        chrom = col_list[0]  # VCF CHROM
        if chrom.lower().startswith("chr"):
            chrom = chrom[3:]
        # if len(chrom) == 2 and chrom[1] not in "0123456789":
        #     chrom = chrom[0]

        pos = int(col_list[1])  # VCF POS
        x = col_list[3]         # VCF REF
        y = col_list[4]         # VCF ALT

        # if multiple REF or ALT alleles are given, ignore mutation entry (could mean seq error, could mean deletion)
        if len(x) != 1:
            N_skipped += 1
            continue
        if len(y) != 1:
            N_skipped += 1
            continue

        raw_mutations.append((chrom, pos, x, y))

    # MAX = 1e7
    if len(raw_mutations) > 0:
        # if len(raw_mutations) > MAX:
        #     raw_mutations = raw_mutations[:MAX]

        contexts = get_context_batch(raw_mutations, asm)
        # print("CONTEXTS", contexts)

        if contexts is None:
            return None, None

        if len(contexts) == 0:
            return None, None

        for (chrom, pos, x, y) in raw_mutations:
            p5, p3 = contexts.get((chrom, pos), ("N", "N"))
            # print("RESULT: {} {}".format(p5, p3))

            if len(set([p5, x, y, p3]) - set(nucleotides)) > 0:
                # print(chrom, pos, p5, p3, x)
                # print("Skipping invalid nucleotides")
                N_skipped += 1
                continue

            if x in "CT":
                mutations[p5 + p3 + x + y] += 1.0
            else:
                # complementary mutation
                mutations[cn[p3] + cn[p5] + cn[x] + cn[y]] += 1.0

    N_loaded = int(sum(mutations.values()))
    processing_stats = {'loaded': N_loaded, 'skipped': N_skipped, 'format': 'VCF'}
    return mutations, processing_stats



import argparse
import signal
import sys

import mutagene
from mutagene.version import __version__

# from mutagene.io.profile import format_profile
# from mutagene.io.mutations_profile import read_VCF_profile
# from mutagene.io.decomposition import write_decomposition

from mutagene.cli.fetch_menu import FetchMenu
from mutagene.cli.profile_menu import ProfileMenu
from mutagene.cli.motif_menu import MotifMenu
from mutagene.cli.signature_menu import SignatureMenu
from mutagene.cli.rank_menu import RankMenu
from mutagene.cli.benchmark_menu import BenchmarkMenu

import logging

logger = logging.getLogger(__name__)

genome_error_message = 'requires genome name argument -g hg19, hg38, mm10, see http://hgdownload.cse.ucsc.edu/downloads.html for more'


class MutaGeneApp(object):
    def __init__(self):
        signal.signal(signal.SIGINT, self.signal_handler)
        # ignore BrokenPipeError: [Errno 32] Broken pipe which occurs when using less or head
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)

        parser = argparse.ArgumentParser(
            prog="mutagene",
            description='MutaGene version {} - Analysis of mutational processes and driver mutations'.format(mutagene.__version__),
            formatter_class=argparse.RawDescriptionHelpFormatter,
            add_help=False,
            # usage="%(prog)s [options]",
            # formatter_class=argparse.RawTextHelpFormatter
        )
        # parser._positionals.title = 'Positional arguments'
        parser._optionals.title = 'Global optional arguments'
        parser.add_argument('-v', '--verbose', action='count', default=0, help='Print additional messages (-v, -vv)')
        parser.add_argument('-V', '--version', action='version', version="%(prog)s " + __version__, help='Show version and exit')
        parser.add_argument('-h', '--help', action='help', default=argparse.SUPPRESS,
                            help='Show this help message and exit')

        subparsers = parser.add_subparsers(
            help="",
            metavar='{fetch, profile, rank, motif, signature}',
            description="""\
        fetch - Load data such as genomes and cancer datasets from demote sources (alias: download)
        profile - Create a mutational profile given a sample with mutations
        rank - Predict driver mutations by ranking observed mutations with respect to their expected mutability
        motif - Test samples for presence of mutational motifs
        signature - Identify activity of existing mutational signatures in samples or derive new signatures (aliases: identify, decompose)\
            """, dest='command', title='Choose MutaGene subpackage')

        parser_mapping = {
            'fetch': {'class': FetchMenu, 'aliases': ['download']},
            'profile': {'class': ProfileMenu, 'aliases': []},
            'rank': {'class': RankMenu, 'aliases': ['driver']},
            'motif': {'class': MotifMenu, 'aliases': []},
            'signature': {'class': SignatureMenu, 'aliases': ['identify', 'decompose']},
            'benchmark': {'class': BenchmarkMenu, 'aliases': []},
        }

        for command, menu in parser_mapping.items():
            # initialize parser object for each subparser
            parser_mapping[command]['parser'] = menu['class'](
                subparsers.add_parser(
                    command,
                    add_help=True,
                    aliases=menu['aliases'],
                    formatter_class=argparse.RawDescriptionHelpFormatter,
                ))

        args = parser.parse_args()

        if not args.command:
            parser.print_help()
            sys.exit(1)

        # LOGGER INIT
        levels = [logging.WARNING, logging.INFO, logging.DEBUG]
        level = levels[min(len(levels) - 1, args.verbose)]  # capped to number of levels
        logging.basicConfig(level=level,
                            format="%(levelname)s %(message)s")

        # Calling callback method for parser object by command name or alias
        parser_class = None
        if args.command in parser_mapping:
            parser_class = parser_mapping[args.command]['parser']
        else:
            for command, mapping in parser_mapping.items():
                if args.command in mapping['aliases']:
                    parser_class = mapping['parser']
                    break

        # should not happen if we have a correct name or an alias for a command
        if not parser_class:
            parser.print_help()
            sys.exit(1)

        parser_class.callback(args)

    @classmethod
    def signal_handler(cls, signal, frame):
        # logger.warning('Interrupted')
        sys.exit(0)


if __name__ == '__main__':
    MutaGeneApp()



import pathlib
import glob
import os
# import pprint
import numpy as np

from math import ceil
from os.path import isfile
from multiprocessing import Pool

from mutagene.io.profile import write_profile, read_profile_file
from mutagene.io.decomposition import write_decomposition, read_decomposition
from mutagene.signatures.identify import decompose_mutational_profile_counts
from mutagene.benchmark.deconstructsigs import deconstruct_sigs, deconstruct_sigs_custom


def gen_benchmark_2combinations(data_root, signature_names, W):
    name_to_idx = {}
    for i, s in enumerate(signature_names):
        name_to_idx[s] = i

    signature_names = list(name_to_idx.keys())
    # print(signature_names)
    # signature_ids = [str(name.split()[1]) for name in signature_names]

    # for r in [0.9, 0.8, 0.7, 0.6, 0.5]:
    for r in [0.7, 0.5]:
        for l in [0.1, 0.2]:
            # for n in [10, 50, 100, 500, 1000, 10000]:
            for n in [50, 500]:
                gen_sample_2combinations(data_root, signature_names, W, ratio=r, noise_level=l, n_mutations=n)


def gen_sample_2combinations(data_root, signature_ids, W, ratio=0.7, noise_level=0.05, n_mutations=10):
    assert 0.0 < ratio < 1.0
    assert 0.0 <= noise_level < ratio
    assert n_mutations > 0
    # np.set_printoptions(precision=4)
    N = W.shape[1]
    assert N >= 5

    n_replica = 10

    for i in range(N):
        for j in range(N):
            # if i == j:
            #     continue

            dirname = "{}/2comb/{}_{}_{}".format(data_root, N, signature_ids[i], signature_ids[j])
            pathlib.Path(dirname).mkdir(parents=True, exist_ok=True)
            print(dirname)

            for r in range(n_replica):
                fname = dirname + "/{:02d}_{:02d}_{:02d}_{:02d}".format(int(ratio * 10), int(noise_level * 100), int(n_mutations), r)
                # print(fname)

                profile_fname = fname + ".profile"
                info_fname = fname + ".info"

                h0 = np.zeros(N)
                h0[i] = ratio
                h0[j] = 1.0 - ratio
                h0 /= h0.sum()

                v0 = W.dot(h0)
                v0_counts = np.random.multinomial(ceil((1.0 - noise_level) * n_mutations), v0 / v0.sum())
                # print(v0_counts)
                v0_counts += np.random.multinomial(ceil(noise_level * n_mutations), [1.0 / v0.shape[0]] * v0.shape[0])  # uniform distribution
                # print(v0_counts)

                write_profile(profile_fname, v0_counts)
                # write_decomposition(info_fname, h0, signature_ids)
                write_decomposition(info_fname, h0, signature_ids, "synthetic", write_zeros=True)


def run_benchmark_2combinations(data_root, N, signature_ids, W, force=False):
    methods = ['MLE', 'MLEZ', 'AICc', 'BIC', 'AICcZ', 'BICZ']

    for fname in glob.glob("{}/2comb/{}_**/*.profile".format(data_root, N), recursive=True):
        print(fname)
        profile = read_profile_file(fname)

        for method in methods:
            info = "{}.{}.info".format(fname.split(".")[0], method)
            if isfile(info) and not force:
                continue

            _, _, results = decompose_mutational_profile_counts(
                profile,
                (W, signature_ids),
                method,
                debug=False,
                others_threshold=0.0)
            exposure_dict = {x['name']: x['score'] for x in results}
            exposure = [exposure_dict[name] for name in signature_ids]
            write_decomposition(info, np.array(exposure), signature_ids)


def run_benchmark_2combinations_deconstruct_sigs_helper(data):
    fname, ds_info, signature_ids, sigtype = data
    print(fname)
    results = deconstruct_sigs_custom(fname, signatures=int(sigtype))
    # print(results)
    exposure_dict = {x['name']: x['score'] for x in results}
    exposure = [exposure_dict[name] for name in signature_ids]
    write_decomposition(ds_info, np.array(exposure), signature_ids)


def run_benchmark_2combinations_deconstruct_sigs(data_root, N, signature_ids, W, force=False):
    def get_iterator():
        for fname in glob.glob("{}/2comb/{}_**/*.profile".format(data_root, N), recursive=True):
            sigtype = fname.split("/")[-2].split("_")[0]
            ds_info = fname.split(".")[0] + ".ds.info"
            if isfile(ds_info) and not force:
                continue
            yield (fname, ds_info, signature_ids, sigtype)

    with Pool(8) as p:
        p.map(run_benchmark_2combinations_deconstruct_sigs_helper, get_iterator(), 100)


def aggregate_benchmarks(data_root):
    methods = {
        "original": ".info",
        "mle": ".MLE.info",
        "mlez": ".MLEZ.info",
        "ds": ".ds.info",
        'aicc': '.AICc.info',
        'bic': '.BIC.info',
        'aiccz': '.AICcz.info',
        'bicz': '.BICz.info',
    }

    # generate full panel for all signatures in 30 x 30 signatures analysis
    with open("{}/2comb/res1-1.txt".format(data_root), 'w') as o:
        o.write("sigtype\tsig1\tsig2\tratio\tnoise\tnmut\treplica\tmethod\tsignature\tvalue\n")
        for fname in glob.glob("data/benchmark/2comb/**/*.profile", recursive=True):
            sigtype, sig1, sig2 = fname.split("/")[-2].split("_")
            ratio, noise, nmut, replica = fname.split("/")[-1].split(".")[0].split("_")

            for method in methods:
                info_fname = fname.split(".")[0] + methods[method]
                values, names = read_decomposition(info_fname)

                if values is None:
                    continue

                for value, signature in zip(values, names):
                    o.write("{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\n".format(sigtype, sig1, sig2, ratio, noise, nmut, replica, method, signature, value))

    # only report the signature 2 value (as in DeconstructSigs benchmark)
    with open("{}/2comb/res2-2.txt".format(data_root), 'w') as o:
        o.write("sigtype\tsig1\tsig2\tratio\tnoise\tnmut\treplica\tmethod\tvalue\n")
        for fname in glob.glob("{}/2comb/**/*.profile".format(data_root), recursive=True):
            sigtype, sig1, sig2 = fname.split("/")[-2].split("_")
            ratio, noise, nmut, replica = fname.split("/")[-1].split(".")[0].split("_")

            for method in methods:
                # if method == 'ds' and sigtype != "30":
                #     continue

                info_fname = fname.split(".")[0] + methods[method]
                values, names = read_decomposition(info_fname)

                if values is None:
                    continue

                for value, signature in zip(values, names):
                    if sig2 == signature:
                        o.write("{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\n".format(sigtype, sig1, sig2, ratio, noise, nmut, replica, method, value))


if __name__ == '__main__':
    pass



import urllib
import math

import numpy as np

from scipy.optimize import minimize
from scipy.optimize import nnls
from scipy.spatial.distance import cosine
# from scipy.optimize import basinhopping  # , differential_evolution
# from scipy.optimize import fmin_cobyla
from scipy.stats import entropy

from mutagene.signatures import get_dummy_signatures_lists

import logging
logger = logging.getLogger(__name__)


def multi_kl(p, q):
    """Kullback-Liebler divergence from multinomial p to multinomial q,
    expressed in bits."""
    # Clip before taking logarithm to avoid NaNs (but still exclude
    # zero-probability mixtures from the calculation)
    return np.sum(p * (np.log2(p.clip(1e-10, 1)) - np.log2(q.clip(1e-10, 1))))


def multi_js(p, q):
    """Jensen-Shannon divergence (symmetric) between two multinomials,
    expressed in bits."""
    # D_{JS}(P\|Q) = (D_{KL}(P\|Q) + D_{KL}(Q\|P)) / 2
    m = 0.5 * (p + q)
    return 0.5 * (multi_kl(p, m) + multi_kl(q, m))
    # return 0.5 * (
    #     (q * (np.log2(q.clip(1e-10, 1)) - np.log2(p.clip(1e-10, 1)))).sum(0) +
    #     (p * (np.log2(p.clip(1e-10, 1)) - np.log2(q.clip(1e-10, 1)))).sum(0))


# Define minimization function
def DivergenceKL(x, A, b):
    b = b + 1e-17
    v = A.dot(x) + 1e-17
    return entropy(v, b)


# Define minimization function
def DivergenceJS(x, A, b):
    b = b + 1e-17
    v = A.dot(x) + 1e-17
    m = (v + b) / 2.0
    return 0.5 * (entropy(v, m) + entropy(b, m))


def Cos(x, A, b):
    """ cosine distance """
    return cosine(A.dot(x), b)


def Elastic(x, A, b):
    """ elastic net """
    alpha = 1.0
    l1_ratio = 0.5

    k = np.count_nonzero(x)  # signatures
    n = b.sum()  # samples

    objective = np.linalg.norm(A.dot(x) - b) / (0.5 * n)
    # l1 reg
    objective += alpha * l1_ratio * k**2
    # l2 reg
    # objective += alpha * (1.0 - l1_ratio) * ||w||^2_2
    return objective


# Define minimization function
def Frobenius(x, A, b):
    return np.linalg.norm(A.dot(x) - b)


# Define minimization function
def FrobeniusZero(x, A, b):
    b_hat = A.dot(x)
    b_hat[b == 0.0] = 0.0
    b_hat /= b_hat.sum()
    error = b - b_hat
    return np.linalg.norm(error)


# def DerNegLogLik(x, A, b):    
#     print(b / A.dot(x))
#     DLL = - np.ma.divide(b, A.dot(x)).filled(0)
#     print(x, DLL)
#     return DLL


def NegLogLik(x, A, b):
    """
    Log Likelihood of a mixture of multinomials
    x = parameters of the mixture
    A = signature x 96 channels
    b = the observed profile (includes counts for 96 channels)
    """

    # if x.sum() - 0.00011 > 1.0:
    #     return 1e6

    if x.sum() > 1.0:
        x /= x.sum()

    LL = np.sum(b * np.ma.log(A.dot(x)).filled(0.0))

    # print("NegLL\t{}\t{}\t{}".format(len(x), "\t".join(map(lambda a: "{:.2f}".format(a), x)), -LL))
    return -LL


def NegLogLikOld(x, A, b):
    """
    Log Likelihood of a mixture of multinomials
    x = parameters of the mixture
    A = signature x 96 channels
    b = the observed profile (includes counts for 96 channels)
    """

    # if x.sum() - 0.00011 > 1.0:
    #     return 1e6

    # if x.sum() > 1.0:
    #     x /= x.sum()

    # Compatibility mode with the website:
    if np.sum(x) > 1.0:
        return 1000

    LL = np.sum(b * np.ma.log(A.dot(x)).filled(0))

    # print("LOG\t{}\t{}\t{}".format(len(x), "\t".join(map(lambda a: "{:.2f}".format(a), x)), LL))
    return -LL


def count_threshold(x, threshold=10e-6):
    """ count the number of above-threshold elements in an array """
    return np.sum(x > threshold)


def AIC(x, A, b):
    """
    AIC https://en.wikipedia.org/wiki/Akaike_information_criterion
    """
    k = count_threshold(x)
    # n = b.sum()
    return 2 * NegLogLik(x, A, b) + 2 * k


def AICc(x, A, b):
    """
    AIC with small sample correction
    """
    k = count_threshold(x)
    n = b.sum()
    if n - k - 1 <= 0:
        return AIC(x, A, b)
    else:
        return AIC(x, A, b) + 2 * k * (k + 1) / (n - k - 1)


def BIC(x, A, b):
    """
    BIC https://en.wikipedia.org/wiki/Bayesian_information_criterion
    """
    k = count_threshold(x)
    n = b.sum()
    # print(2 * NegLogLik(x, A, b) + k * np.log(n))
    return 2 * NegLogLik(x, A, b) + k * np.log(n)


IDENTIFY_MIN_FUNCTIONS = {
    'frobenius': Frobenius,
    'frobeniuszero': FrobeniusZero,
    'cos': Cos,
    'elastic': Elastic,
    'kl': DivergenceKL,
    'divergencekl': DivergenceKL,
    'js': DivergenceJS,
    'divergencejs': DivergenceJS,
    'mle': NegLogLik,  # MLE maximizes LogLik or minimizes NegLogLik
    'compat': NegLogLikOld,  # MLE with added context-independent signatures (old compatibility mode)
    'aicc': AICc,  # AIC corrected for small samples
    'bic': BIC,  # BIC
    'mlez': NegLogLik,  # MLE with added context-independent signatures (different name Z given for benchmarking)
    'aiccz': AICc,  # AIC corrected for small samples with added context-independent signatures (different name Z given for benchmarking)
    'bicz': BIC,  # BIC with added context-independent signatures (different name Z given for benchmarking)
}


def get_fingerprint_url(a):
    data = {"s{}".format(i): v for i, v in enumerate(a)}
    return urllib.parse.urlencode(data)


def decompose_mutational_profile_counts(profile, signatures, func="Frobenius", others_threshold=0.05, global_optimization=None, enable_dummy=None):

    config = {
        'enable_dummy': True,
        'show_dummy': False,
        'show_dummy_aggregate': True,
        'global_optimization': False,
    }

    if global_optimization is not None:
        config['global_optimization'] = global_optimization

    if enable_dummy is not None:
        config['enable_dummy'] = enable_dummy
    else:
        if func.lower().endswith('z'):
            config['enable_dummy'] = True

    W = []
    results = []

    W, signature_names = list(signatures)
    signature_names = signature_names[:]  # make a copy
    for name in signature_names:
        results.append({
            'accession': 0,
            'id': 0,
            'pid': 0,
            'name': name,
            'annotation': '',
            'score': 0.0,
            'mutations': 0
        })

    # add dummy signatures
    if config['enable_dummy']:
        for i, (name, values) in enumerate(get_dummy_signatures_lists()):
            dummy = "d" + str(i)
            signature_names.append(dummy)
            W = np.append(W.T, np.array([values, ]), axis=0).T
            results.append({
                'accession': 0,
                'profile': get_fingerprint_url(values),
                'id': 0,
                'pid': 0,
                'name': dummy,
                'annotation': 'Dummy ' + name,
                'score': 0.0,
                'mutations': 0
            })
    # print(W)
    # print(signature_names)

    v = np.array([profile]).ravel()
    v_freq = v / v.sum()

    # Initial guess with NNLS:
    # DATA, residuals = nnls(W.T, X.T.ravel())
    h0, rnorm = nnls(W, v_freq)
    if h0.sum() > 1.0:
        h0 = h0.ravel() / h0.sum()

    # not sure if we need this failsafe option:
    # if np.isnan(h0.sum()):
    #     h0 = np.ones(h0.shape[0]) / h0.shape[0]

    logger.debug("h0 {}".format(h0))

    min_func = IDENTIFY_MIN_FUNCTIONS.get(func.lower(), Frobenius)

    # if debug:
    #     np.set_printoptions(precision=4)
    #     print("--------------------------------------\n")
    #     print("Signature", signatures)
    #     print("NNLS", h0)
    #     print("FRO", round(Frobenius(h0, W, v), 4), "DIV", round(DivergenceKL(h0, W, v), 4))

    # Define constraints and bounds
    # constraints = {'type': 'eq', 'fun': lambda x: np.sum(x) - 1}
    constraints = {'type': 'ineq', 'fun': lambda x: 1.0 - np.sum(x)}
    # bounds = [[0., None],[0., None],[0., None]]

    # bounds = [[0., 1.0] for _ in range(h0.shape[0])]
    bounds = [[0.0, 1.0] for _ in range(h0.shape[0])]

    # v == counts
    # v_freq == frequency (normalized counts)
    if min_func in (NegLogLik, AIC, AICc, BIC):
        v_target = v
    else:
        v_target = v_freq

    def print_fun(x, f, accepted):
        print("{} at minimum {} accepted {}".format(x, f, accepted))

    class RandomDisplacementBounds(object):
        """random displacement with bounds"""

        def __init__(self, xmin=0.0, xmax=1.0, stepsize=0.3):
            self.xmin = xmin
            self.xmax = xmax
            self.stepsize = stepsize

        def __call__(self, x):
            """take a random step but ensure the new position is within the bounds"""
            if np.any(np.isnan(x)):
                return x
            while True:
                xnew = x + np.random.uniform(-self.stepsize, self.stepsize, x.shape)
                xnew = np.where(xnew < 0.0, 0.0, xnew)
                if xnew.sum() > 1.0:
                    xnew += 1.0 - xnew.sum()
                xnew = np.abs(xnew)
                if np.all(xnew <= self.xmax) and np.all(xnew >= self.xmin):
                    break
            return xnew

    # class RandomDisplacementBounds(object):
    #     """random displacement with bounds"""
    #     def __init__(self, xmin, xmax, stepsize=0.1):
    #         self.xmin = xmin
    #         self.xmax = xmax
    #         self.stepsize = stepsize

    #     def __call__(self, x):
    #         """take a random step but ensure the new position is within the bounds"""
    #         while True:
    #             # xnew = x + np.random.uniform(-self.stepsize, self.stepsize, x.shape)
    #             inew = np.random.randint(0, x.shape[0])
    #             xnew = x.copy()
    #             xnew[inew] += np.random.uniform(-self.stepsize, self.stepsize)
    #             # print("it", np.sum(xnew), xnew)
    #             if np.all(xnew <= self.xmax) and np.all(xnew >= self.xmin) and np.sum(xnew) <= 1.0:
    #                 break
    #         return xnew

    # define the new step taking routine and pass it to basinhopping
    take_step = RandomDisplacementBounds()

    def accept_test(f_new, x_new, f_old, x_old):
        if np.any(np.isnan(x_new)):
            return False
        if x_new.sum() > 1.0:
            return False
        if np.any(x_new > 1.0) or np.any(x_new < 0.0) or np.any(np.isnan(x_new)):
            return False
        return True

    def get_constraint(i, upper=True):
        def a(x):
            return 1.0 - x[i]

        def b(x):
            return x[i] - 0.0

        return a if upper else b

    cobyla_constraints = [
        {'type': 'ineq', 'fun': lambda x: 1.0 - np.sum(x)}  # total upper
    ]
    for i in range(h0.shape[0]):
        cobyla_constraints.append(
            {'type': 'ineq', 'fun': get_constraint(i, True)}  # , 'jac': get_constraint(i, True)}  # upper
        )
        cobyla_constraints.append(
            {'type': 'ineq', 'fun': get_constraint(i, False)}  # , 'jac': get_constraint(i, False)}  # lower
        )

    ##############################################################################
    if config['global_optimization']:
        # L-BFGS-B COBYLA SLSQP
        """
        minout = basinhopping(NegLogLik, h0, minimizer_kwargs={  # 'jac': DerNegLogLik,
            'args': (W, v_target), 'method': 'SLSQP',
            'constraints': cobyla_constraints, 'options': {'disp': False, 'ftol': 0.00001, 'eps': 0.0000001}}, niter=100, T=0.001, take_step=take_step, accept_test=accept_test, callback=print_fun)  #, take_step=take_step , callback=print_fun)

        if debug:
            print(minout, type(minout))
            print(minout.message)
            # print("Status: # failures", minout.res.minimization_failures)

        if np.any(np.isnan(minout.x)):
            print("FAILED!!!!!!!!!!!!!!!!")
            if debug:
                print("MINIMIZATION FAILED:", minout.message, minout.nit)
            # fallback decomposition:
            h = h0.ravel() / h0.sum()
            # h = h0.ravel() * 0.0
        else:
            h = minout.x
            if debug:
                print("MAX LIK", h, round(-NegLogLik(h, W, v_target), 4))
                print("MIN FRO", h, round(min_func(h, W, v_target), 4))
                print("FRO", round(Frobenius(h, W, v_target), 4), "DIV", round(DivergenceKL(h, W, v), 4))
        """
        # Bayesian Optimization
        from bayes_opt import BayesianOptimization
        from functools import partial

        optimizer = BayesianOptimization(
            f=partial(min_func, A=W, b=v_target),
            # pbounds=pbounds,
            verbose=2,  # verbose = 1 prints only when a maximum is observed, verbose = 0 is silent
            random_state=1,
        )

    ##############################################################################
    # debug = True
    if not config['global_optimization']:
        minout = minimize(
            min_func, h0, args=(W, v_target),
            method='SLSQP',
            bounds=bounds, constraints=constraints,
            options={'maxiter': 500}
        )

        if minout.success:
            logger.debug("MINIMIZATION: {} {}".format(minout.message, minout.nit))
            h = minout.x
            logger.debug("MAX LIK {} {}".format(h, round(-NegLogLik(h, W, v_target), 4)))
            # print("LIK", round(-NegLogLik(h, W, v), 4), "DIV", round(DivergenceKL(h, W, v), 4))
        else:
            logger.debug("MINIMIZATION FAILED:{} {}".format(minout.message, minout.nit))
            # Minimization did not converge
            # Use our initial guess, but normalize it:
            h = h0.ravel() / h0.sum()

    ##############################################################################
    N_mutations = int(math.ceil(v.sum()))

    for i in range(len(results)):
        results[i]['score'] = h[i]
        results[i]['mutations'] = round(N_mutations * h[i])

    # remove dummy signatures if enabled
    if config['enable_dummy']:
        results = results[:-6]
        signature_names = signature_names[:-6]

    # residuals = min_func(h, W, v)
    reconstructed_profile = W.dot(h)
    ##############################################################################

    below_threshold = []
    above_threshold = []
    # import operator
    # for r in results:
    # print(results)
    j = 0
    for r in sorted(results, key=lambda item: item['score'], reverse=True):
        if others_threshold > 0.0 and round(r['score'], 2) <= others_threshold:
            below_threshold.append(r)
        else:
            j += 1
            r['id'] = j
            above_threshold.append(r)
    results = above_threshold

    # if len(below_threshold) == 1:
    #     # print("ONLY ONE")
    #     results.append(below_threshold[0])
    #     below_threshold = []

    # sum up other signatures
    other_signatures = 0.0
    for r in below_threshold:
        other_signatures += r['score']

    if round(other_signatures, 2) > 0.0:
        results.append({
            'accession': 0,
            'id': 100,
            'pid': 0,
            'name': 'Other signatures',
            'annotation': 'Signatures with individual contrubution &le; 0.05',
            'profile': '',
            'score': other_signatures,
            'mutations': round(N_mutations * other_signatures)
        })
        for j, r in enumerate(below_threshold):
            if round(r['score'], 2) > 0.0:
                r['id'] = 100 + j + 1
                r['pid'] = 100
                results.append(r)

    ll = -NegLogLik(h, W, v_target)
    results.append({
        'accession': 0,
        'id': 201,
        'pid': 0,
        'name': 'LogLik',
        'annotation': '',
        'profile': '',
        'mutations': '',
        'score': ll,
    })

    frobenius = Frobenius(h, W, v_freq)
    results.append({
        'accession': 0,
        'id': 202,
        'pid': 0,
        'name': 'Frobenius',
        'annotation': '',
        'profile': '',
        'mutations': '',
        'score': frobenius,
    })

    frobeniuszero = FrobeniusZero(h, W, v_freq)
    results.append({
        'accession': 0,
        'id': 203,
        'pid': 0,
        'name': 'FrobeniusZero',
        'annotation': '',
        'profile': '',
        'mutations': '',
        'score': frobeniuszero,
    })

    divergencejs = DivergenceJS(h, W, v_freq)
    results.append({
        'accession': 0,
        'id': 204,
        'pid': 0,
        'name': 'DivergenceJS',
        'annotation': '',
        'profile': '',
        'mutations': '',
        'score': divergencejs,
    })

    divergencekl = DivergenceKL(h, W, v_freq)
    results.append({
        'accession': 0,
        'id': 205,
        'pid': 0,
        'name': 'DivergenceKL',
        'annotation': '',
        'profile': '',
        'mutations': '',
        'score': divergencekl,
    })

    summary = []
    summary.append({
        'accession': 0,
        'name': 'Query profile',
        'annotation': 'Query profile derived from mutations uploaded by the user',
        'profile': get_fingerprint_url(profile)
    })

    summary.append({
        'accession': 0,
        'name': 'Reconstructed profile',
        'annotation': 'Profile represented as a combination of signatures in the table below',
        'profile': get_fingerprint_url(reconstructed_profile)
    })

    return h, summary, results



import sys
import os
# from pathlib import Path

# from mutagene.mutability.mutability import rank, THRESHOLD_DRIVER, THRESHOLD_PASSENGER
# from mutagene.io.cohorts import read_cohort_mutations_from_tar
# from mutagene.io.cohorts import read_cohort_size_from_profile_file, list_cohorts_in_tar
# from mutagene.io.profile import read_profile_file
# from mutagene.io.protein_mutations_MAF import read_MAF_with_genomic_context

from mutagene.benchmark.generate_benchmark import (
    gen_benchmark_2combinations,
    run_benchmark_2combinations,
    run_benchmark_2combinations_deconstruct_sigs,
    aggregate_benchmarks
)

from mutagene.benchmark.multiple_benchmark import (
    multiple_benchmark,
    multiple_benchmark_run,
    aggregate_multiple_benchmarks
)

from mutagene.io.profile import read_signatures

import logging
logger = logging.getLogger(__name__)


class BenchmarkMenu(object):
    def __init__(self, parser):
        required_group = parser.add_argument_group('Required arguments')
        required_group.add_argument("--mode", "-m", choices=[
            'pairwise_gen',
            'pairwise_run',
            'pairwise_run_ds',
            'multiple_gen',
            'multiple_run',
            'multiple_run_ds',
            'aggregate'], help="Multiple or pairwise mode, etc", type=str)
        required_group.add_argument("--signatures", "-i", nargs='*', help="Signatures (5, 10, 30, ...), default 30", type=str, default=["30"])

        # dirname = os.path.dirname(os.path.realpath(__file__))
        # default_root = dirname + "/../data/benchmark"
        default_root = "data/benchmark"
        default_root = os.path.normpath(default_root)

        required_group.add_argument(
            "--root",
            help="path to benchmark data, default {}".format(default_root),
            type=str,
            default=default_root)

        # required_group.add_argument('--genome', "-g", help="Location of genome assembly file in 2bit format", type=str)
        self.parser = parser
        pass

    def callback(self, args):
        if args.mode.startswith("pair"):
            for i in args.signatures:
                i = int(i)
                W, signature_names = read_signatures(i)

                if args.mode.endswith('gen'):
                    gen_benchmark_2combinations(args.root, signature_names, W)
                elif args.mode.endswith('run'):
                    run_benchmark_2combinations(args.root, i, signature_names, W, force=True)
                elif args.mode.endswith('run_ds'):
                    run_benchmark_2combinations_deconstruct_sigs(args.root, i, signature_names, W, force=True)

        elif args.mode.startswith('multiple'):
            if args.mode.endswith('gen'):
                pass
            elif args.mode.endswith('run'):
                multiple_benchmark_run(i, signature_names, W, force=True)
            elif args.mode.endswith('run_ds'):
                pass
            multiple_benchmark()
            aggregate_multiple_benchmarks()

        elif args.mode == 'aggregate':
            aggregate_benchmarks(args.root)

        else:
            print("Unknown benchmark action mode")
            self.parser.print_usage()
            sys.exit(1)



import argparse
import sys
import logging

from tqdm import tqdm

from mutagene.io.profile import read_signatures
from mutagene.profiles.profile import get_multisample_mutational_profile
from mutagene.profiles.profile import generate_resampled_profiles
from mutagene.io.context_window import read_mutations
from mutagene.signatures.identify import decompose_mutational_profile_counts
from mutagene.signatures.identify import IDENTIFY_MIN_FUNCTIONS
from mutagene.io.decomposition import write_decomposition


logger = logging.getLogger(__name__)

genome_error_message = """requires genome name argument -g hg19, hg38, mm10, see http://hgdownload.cse.ucsc.edu/downloads.html for more
                          Use mutagene fetch to download genome assemblies"""


class SignatureMenu(object):
    def __init__(self, parser):
        required_group = parser.add_argument_group('Required arguments')
        required_group.add_argument("--infile", "-i", help="Input file in VCF or MAF format", type=argparse.FileType('r'))
        required_group.add_argument('--genome', "-g", help="Location of genome assembly file in 2bit format", type=str, default='hg19')
        required_group.add_argument(
            "--signatures", "-s",
            choices=["5", "10", "30", "49", "53", "MGA", "MGB", "COSMICv2", "COSMICv3", "KUCAB"],
            help="Collection of signatures to use", type=str, default="COSMICv2")

        optional_group = parser.add_argument_group('Optional arguments')
        optional_group.add_argument('--input-format', "-f", help="Input format: MAF, VCF", type=str, choices=['MAF', 'VCF', 'TCGI'], default='MAF')
        optional_group.add_argument(
            '--outfile', "-o", nargs='?', type=argparse.FileType('w'), default=sys.stdout,
            help="Name of output file, will be generated in TSV format")

        # for backwards compatibility with 0.8.X add a hidden action that would just take anything as a valid input
        optional_group.add_argument('action', nargs="?", metavar="")

        advanced_group = parser.add_argument_group('Advanced arguments')
        advanced_group.add_argument('--method', "-m", help="Method defines the function minimized in the optimization procedure", type=str, default='MLE', nargs='?')
        advanced_group.add_argument('--no-unexplained-variance', "-U", help="Do not account for unexplained variance (non-context dependent mutational processes and unknown signatures)", action='store_false')
        advanced_group.add_argument('--mutations-threshold', "-t", help="Only report signatures with mutations above the threshold", type=int, default=0)
        advanced_group.add_argument('--keep-only', "-k", help="Keep only the signatures in the list, separated by commas e.g. 1,3,5", type=str, default=None)

        bootstrap_group = parser.add_argument_group('Bootstrap-specific arguments')
        bootstrap_group.add_argument('--bootstrap', "-b", help="Use the bootstrap to calculate confidence intervals", action='store_true')
        bootstrap_group.add_argument('--bootstrap-replicates', "-br", help="Number of bootstrap replicates", type=int, default=100)
        bootstrap_group.add_argument('--bootstrap-confidence-level', "-bcl", help="Confidence level", type=int, default=95)
        bootstrap_group.add_argument('--bootstrap-method', "-bm", help="Bootstrap method (t: t-distribution, p: percentile)", type=str, choices=['t', 'p'], default='p')
        self.parser = parser

    def identify(self, args):
        if not args.infile:
            logger.warning("Provide input file in VCF or MAF format (-i) and a corresponding genome assembly (-g)")
            return
        if not args.genome:
            logger.warning(genome_error_message)
            return
        if not args.signatures:
            logger.warning("Set of signatures required. Use 5 and 10 for MUTAGENE-5 and MUTAGENE-10. Use 30 for COSMIC-30")
            return

        if args.method.lower() not in IDENTIFY_MIN_FUNCTIONS:
            logger.warning('Unknown method provided')
            return

        if args.bootstrap_replicates < 10:
            logger.warning("Number of bootstrap replicates too low. Specify at least 10 replicates")
            return

        if args.bootstrap_confidence_level < 70:
            logger.warning("Specify confidence level of at least 70% and less than 99%")
            return

        only = None
        if args.keep_only is not None:
            only = args.keep_only.split(",")
            if len(only) < 1:
                logger.warning("List of signatures to keep for the analysis is empty")
                return
            logger.warning("We will only analyze signatures in this list: {}".format(", ".join(only)))

        if 'input_format' not in args:
            # guess format from file name
            name = args.infile.name.upper()
            if name.endswith("MAF"):
                args.input_format = "MAF"
            elif name.endswith("VCF"):
                args.input_format = "VCF"
            else:
                logger.warning("Input format was not specified. Assuming it is MAF")
                args.input_format = "MAF"

        W, signature_names = read_signatures(args.signatures, only=only)

        try:
            mutations, _, processing_stats = read_mutations(args.input_format, args.infile, args.genome, window_size=1)
        except Exception as e:
            e_message = getattr(e, 'message', repr(e))
            logger.warning(
                "Parsing {0} failed. "
                "Check that the input file is in {0} format "
                "or specify a different format using option -f \n"
                "{1}".format(args.input_format, e_message))

            if logger.root.level == logging.DEBUG:
                raise
            return

        samples_profiles = get_multisample_mutational_profile(mutations, counts=True)
        samples_results = {}

        for sample, profile in samples_profiles.items():
            _, _, results = decompose_mutational_profile_counts(
                profile,
                (W, signature_names),
                args.method,
                others_threshold=0.0,
                enable_dummy=args.no_unexplained_variance)
            samples_results[sample] = results

        if not args.bootstrap:
            write_decomposition(args.outfile, samples_results, signature_names, mutations_threshold=args.mutations_threshold)
        else:
            bootstrap_samples_results = {}
            for sample, profile in samples_profiles.items():
                bootstrap_results = []
                for resampled_profile in tqdm(generate_resampled_profiles(profile, args.bootstrap_replicates), total=args.bootstrap_replicates):
                    _, _, results = decompose_mutational_profile_counts(
                        resampled_profile,
                        (W, signature_names),
                        args.method,
                        others_threshold=0.0,
                        enable_dummy=args.no_unexplained_variance)
                    bootstrap_results.append(results)
                bootstrap_samples_results[sample] = bootstrap_results

            write_decomposition(
                args.outfile, samples_results, signature_names,
                mutations_threshold=args.mutations_threshold,
                bootstrap_method=args.bootstrap_method,
                profile=samples_profiles,
                bootstrap_results=bootstrap_samples_results,
                bootstrap_level=args.bootstrap_confidence_level)

    def callback(self, args):
        self.identify(args)



import argparse
import sys
import logging
from pathlib import Path

from mutagene.mutability.mutability import rank, THRESHOLD_DRIVER, THRESHOLD_PASSENGER
from mutagene.io.cohorts import read_cohort_mutations_from_tar
from mutagene.io.cohorts import read_cohort_size_from_profile_file, list_cohorts_in_tar
from mutagene.io.profile import read_profile_file
from mutagene.io.context_window import read_mutations
from mutagene.profiles.profile import get_pooled_multisample_mutational_profile
from mutagene.io.protein_mutations_MAF import read_protein_mutations_MAF


logger = logging.getLogger(__name__)
genome_error_message = """requires genome name argument -g hg19, hg38, mm10, see http://hgdownload.cse.ucsc.edu/downloads.html for more
                          Use mutagene fetch to download genome assemblies"""


class RankMenu(object):
    def __init__(self, parser):
        required_group = parser.add_argument_group('Required arguments')
        required_group.add_argument("--infile", "-i", help="Input file in MAF format", type=argparse.FileType('r'))
        required_group.add_argument('--genome', "-g", help="Location of genome assembly file in 2bit format", type=str, default='hg19')
        # parser.add_argument('--mode', "-m", help="n (nucleotide) or aa (amino acid) mode", type=str, default="aa")

        optional_group = parser.add_argument_group('Optional arguments')
        optional_group.add_argument('--outfile', "-o", nargs='?', type=argparse.FileType('w'), default=sys.stdout)
        # Suppressed with issue #51
        optional_group.add_argument('--input-format', "-f", help=argparse.SUPPRESS, type=str, choices=['MAF', 'VCF', 'TCGI'], default='MAF')  # "Input format: MAF, VCF, or TCGI"
        optional_group.add_argument('--cohorts-file', type=str, help="Location of tar.gz container or directory for cohorts", default="cohorts.tar.gz", nargs='?')
        optional_group.add_argument('--cohort', "-c", type=str, help="Name of cohort with observed mutations", nargs='?', const="")

        advanced_group = parser.add_argument_group('Advanced arguments')
        advanced_group.add_argument('--profile', "-p", help="Override profile to calculate mutability, may also describe cohort size", type=str)
        advanced_group.add_argument('--nsamples', "-n", type=int, help="Override cohort size")
        advanced_group.add_argument('--threshold-driver', "-td", help="BScore threshold between Driver and Pontential Driver mutations", type=float, default=THRESHOLD_DRIVER)
        advanced_group.add_argument('--threshold-passenger', "-tp", help="BScore threshold between Pontential Driver and Passenger mutations", type=float, default=THRESHOLD_PASSENGER)

    def callback(self, args):
        """ Rank mutations """

        # no need to read mutations if there is a simple bail out
        if args.cohort is not None:
            if not Path(args.cohorts_file).is_file():
                logger.warning("Cohorts file missing. Download with \"mutagene fetch_cohorts\"")
                return
            if len(args.cohort) == 0:
                logger.warning("List of available cohorts:\n" + list_cohorts_in_tar(args.cohorts_file))
                return

        try:
            mutations, _, processing_stats = read_mutations(args.input_format, args.infile, args.genome, window_size=1)
        except Exception as e:
            e_message = getattr(e, 'message', repr(e))
            logger.warning(
                "Parsing {0} failed. "
                "Check that the input file is in {0} format "
                "or specify a different format using option -f \n"
                "{1}".format(args.input_format, e_message))

            if logger.root.level == logging.DEBUG:
                raise
            return

        msg = "Loaded {} mutations".format(processing_stats['loaded'])
        if processing_stats['skipped'] > 0:
            msg += " skipped {} mutations".format(processing_stats['skipped'])
        logger.info(msg)

        if not len(mutations) or not len(mutations[list(mutations.keys())[0]]):
            logger.warning('No mutations to rank. Check that the input file is in MAF format and correct genome assembly is chosen')
            return

        # read protein mutatations:
        args.infile.seek(0)
        protein_mutations, processing_stats = read_protein_mutations_MAF(args.infile, args.genome)
        msg = "Loaded {} protein mutations in {} samples".format(processing_stats['loaded'], processing_stats['nsamples'])
        if processing_stats['skipped'] > 0:
            msg += " skipped {} protein mutations".format(processing_stats['skipped'])
        logger.info(msg)

        cohort_size = len(mutations.keys())
        logger.info("Cohort size: {}".format(cohort_size))
        profile = get_pooled_multisample_mutational_profile(mutations, counts=True)
        cohort_aa_mutations = None

        # optionals:
        # overriding profile, cohort size and observed mutations based on precalc cohort
        if args.cohort is not None:
            profile, cohort_size, cohort_aa_mutations, cohort_na_mutations = read_cohort_mutations_from_tar(args.cohorts_file, args.cohort)
            logger.info("Precalculated cohort and profile loaded")
            logger.info("Cohort size: {}".format(cohort_size))

        # overriding profile:
        if args.profile:
            profile = read_profile_file(args.profile)
            if profile:
                logger.info('Profile overridden')
            else:
                logger.info('Profile could not be loaded')
                return
            cohort_size_new = read_cohort_size_from_profile_file(args.profile)
            if cohort_size_new:
                cohort_size = cohort_size_new
                logger.info('Cohort size loaded from profile N=' + str(cohort_size))

        # overriding n samples
        if args.nsamples:
            cohort_size = args.nsamples
            logger.info('Cohort size overridden N=' + str(cohort_size))

        logger.info("THRESHOLD_DRIVER: {}".format(args.threshold_driver))
        logger.info("THRESHOLD_PASSENGER: {}".format(args.threshold_passenger))

        # ranking:
        rank(protein_mutations, args.outfile, profile, cohort_aa_mutations, cohort_size, args.threshold_driver, args.threshold_passenger)



import setuptools
from mutagene.version import __version__

setuptools.setup(name='mutagene',
                 version=__version__,
                 description='Mutational analysis with Python',
                 long_description=open('README.md').read().strip(),
                 long_description_content_type="text/markdown",
                 author='Alexander Goncearenco',
                 author_email='alexandr.goncearenco@nih.gov',
                 url='http://www.ncbi.nlm.nih.gov/projects/mutagene/',
                 install_requires=[
                        'bayesian-optimization',
                        'clustergrammer',
                        'numpy>=1.18',
                        'pandas>=0.25',
                        'requests',
                        'scikit-learn>=0.22',
                        'scipy>=1.4',
                        'statsmodels',
                        'pyvcf',
                        'twobitreader',
                        'tqdm'],
                 test_suite='nose.collector',
                 tests_require=[
                     'nose',
                     'pytest',
                     'pytest-cov',
                     'coverage'],
                 scripts=['bin/mutagene'],
                 # package_dir={'mutagene': 'src'},
                 packages=setuptools.find_packages(exclude=["docs", "contrib", "*.tests", "*.tests.*", "tests.*", "tests"]),
                 include_package_data=True,
                 license='Public Domain',
                 zip_safe=False,
                 keywords='bioinformatics cancer mutations',
                 classifiers=[
                     'Programming Language :: Python :: 3 :: Only',
                     "Operating System :: OS Independent",
                     'License :: Public Domain',
                     'Development Status :: 4 - Beta',
                     'Intended Audience :: Science/Research',
                     'Topic :: Scientific/Engineering :: Medical Science Apps.',
                     'Environment :: Console'
                 ])
